
# Module colormap for creating custom color maps.

# Note that many colormaps are also defined in matplotlib and can be set by
#   >>> from matplotlib import cm
#   >>> colormap = cm.get_cmap('Greens')
# for example, to get colors ranging from white to green.
# See matplotlib._cm for the data defining various maps.

#-------------------------
def make_colormap(colors):
#-------------------------
    """
    Define a new color map based on values specified in the dictionary
    colors, where colors[z] is the color that value z should be mapped to,
    with linear interpolation between the given values of z.

    The z values (dictionary keys) are real numbers and the values
    colors[z] can be either an RGB list, e.g. [1,0,0] for red, or an
    html hex string, e.g. "#ff0000" for red.
    """

    from matplotlib.colors import LinearSegmentedColormap, ColorConverter
    from numpy import sort
    
    z = sort(colors.keys())
    n = len(z)
    z1 = min(z)
    zn = max(z)
    x0 = (z - z1) / (zn - z1)
    
    CC = ColorConverter()
    R = []
    G = []
    B = []
    for i in range(n):
        #i'th color at level z[i]:
        Ci = colors[z[i]]      
        if type(Ci) == str:
            # a hex string of form '#ff0000' for example (for red)
            RGB = CC.to_rgb(Ci)
        else:
            # assume it's an RGB triple already:
            RGB = Ci
        R.append(RGB[0])
        G.append(RGB[1])
        B.append(RGB[2])

    cmap_dict = {}
    cmap_dict['red'] = [(x0[i],R[i],R[i]) for i in range(len(R))]
    cmap_dict['green'] = [(x0[i],G[i],G[i]) for i in range(len(G))]
    cmap_dict['blue'] = [(x0[i],B[i],B[i]) for i in range(len(B))]
    mymap = LinearSegmentedColormap('mymap',cmap_dict)
    return mymap

def showcolors(cmap):
    from pylab import colorbar, clf, axes, linspace, pcolor, meshgrid, show
    #from scitools.easyviz.matplotlib_ import colorbar, clf, axes, linspace,\
                 #pcolor, meshgrid, show, colormap
    clf()
    x = linspace(0,1,21)
    X,Y = meshgrid(x,x)
    pcolor(X,Y,0.5*(X+Y), cmap=cmap, edgecolors='k', shading='faceted')
    #colormap(cmap)
    #X,Y = meshgrid(x,x,indexing='ij')
    #pcolor(X,Y,0.5*(X+Y), colormap=cmap, edgecolors='k', shading='faceted')
    colorbar()
    #show()

def ryb():
    """
    Red-yellow-blue colormap
    """
    colors = {0.0: '#ff0000',\
              0.5: '#ffff00',\
              1.0: '#0000ff'}
    colormap = make_colormap(colors)
    return colormap

def byr():
    """
    Blue-yellow-red colormap
    """
    colors = {0.0: '#0000ff',\
              0.5: '#ffff00',\
              1.0: '#ff0000'}
    colormap = make_colormap(colors)
    return colormap

def yrb():
    """
    Yellow-red-blue colormap
    """
    colors = {0.0: '#ffff00',\
              0.5: '#ff0000',\
              1.0: '#0000ff'}
    colormap = make_colormap(colors)
    return colormap

def rwb():
    """
    Red-white-blue colormap
    """
    colors = {0.0: '#ff0000',\
              0.5: '#ffffff',\
              1.0: '#0000ff'}
    colormap = make_colormap(colors)
    return colormap

def white():
    """
    All white colormap
    """
    colors = {0.0: '#ffffff',\
              1.0: '#ffffff'}
    colormap = make_colormap(colors)
    return colormap

