# Trimmed down version without solvers.
# For use only with fortran version of the code.
# Modified by RJL for sphere paper.


#!/usr/bin/env python
# encoding: utf-8
#  ======================================================================
#  Package:     pyclaw
#  File:        controller.py
#  Created:     Feb 19, 2008
#  Author:      Kyle Mandli
#  ======================================================================
"""
Module for the ClawController class to control a computation and plotting
of results.


"""

import pdb

import logging
import sys
import os
import copy

from error import ClawError
from data import ClawRunData, ClawPlotData, ClawPlotItem
from solution import ClawSolution
from util import ClawFrameCounter

from numpy.lib.function_base import linspace

class ClawController(object):
    """Main pyclaw class to control a computation and plotting of results

    Members:
        Attributes:
          Global information for running and/or plotting
        
          Executable support (runxclaw routine support):
            xdir = Executable path
            rundir = Directory to run from (containing *.data files)
            xclawout = where to write timestep messages
            xclawerr = where to write error messages
            runmake = run make in xdir before xclawcmd
            savecode = save a copy of *.f files in outdir
            
          Solver attributes:
            solver = Solver object
        
          Output parameters:
            outdir = Output directory
            overwrite = ok to overwrite old results in outdir?
            keep_copy = Keep a copy of every output time
            write_aux_init = Write out initial auxillary array
            write_aux_always = Write out auxillary array at every time step
            output_format = Format to output the data in
            output_file_prefix = File prefix to be appended to output files
            output_options = Output options passed to function writing and 
                reading data in output_format's format
                
        plotdata = An instance of a ClawPlotData object defining the objects
            plot parameters
        rundata = An instance of a ClawRunData object defining the objects
            run parameters
        
    """
    #  ======================================================================
    #   Initialization routines
    #  ======================================================================
    def __init__(self):
        """
        Initialization routine for a ClawController object.
        """
        
        self.viewable_attributes = ['xdir','rundir','outdir','overwrite',
                        'xclawcmd','xclawout','xclawerr','runmake','savecode',
                        'solver','keep_copy','write_aux_init',
                        'write_aux_always','output_format',
                        'output_file_prefix','output_options']

        # Global information for running and/or plotting
        self.xdir = os.getcwd()         # executes xclawcmd in xdir
        self.rundir = os.getcwd()       # uses *.data from rundir
        self.outdir = os.getcwd()       # directs fort.* to outdir
        self.overwrite = True           # ok to overwrite old results in outdir?

        self.xclawcmd = 'xclaw'         # command to execute (if using fortran)
        if sys.platform == 'cygwin':
             self.xclawcmd = 'xclaw.exe'

        self.xclawout = None            # where to write timestep messages
        self.xclawerr = None            # where to write error messages
        self.runmake = True             # run make in xdir before xclawcmd
        self.savecode = False           # save a copy of *.f files in outdir
        
        # Solver information
        self.solver = None               # Solver object
        self.keep_copy = False           # Keep a copy of every output time
        self.write_aux_init = False      # Write out initial auxillary array
        self.write_aux_always = False    # Write out aux array for all frames
        self.output_format = 'ascii'     # Solver output format
        self.output_file_prefix = None   # Solver output file prefix
        self.output_options = {}         # Output format options
        self.frames = []                 # Saved frames from solver

        # Data objects
        self.rundata = None
        self.plotdata = None
    # ========================================================================
    #   Access methods
    # ======================================================================== 
    def __str__(self):
        """Representation string of the ClawController object"""
        
        output = "ClawController attributes:\n"
        for attr in self.viewable_attributes:
            value = eval('self.%s' % attr)
            output = output + "  %s = %s \n" % (attr,value)
        output = output + '\n'
        if self.rundata is not None:
            output = output + "  Run Data "+str(self.rundata)+"\n"
        if self.plotdata is not None:
            output = output + "  Run Data "+str(self.plotdata)+"\n"
        if self.solver is not None:
            output = output + "  Solver "+str(self.solver)+"\n"
        if len(self.frames) > 0:
            output = output + "  Frames \n"
            for frame in self.frames:
                output = output + "    " + str(frame) + "\n"
        return output
        
    # Override the general access methods to provide solutions access for 
    # convenience
    def __getattribute__(self,name):
        if name is 'solutions':
            return eval('self.solver.solutions')
        return object.__getattribute__(self,name)
            
    def __setattr__(self,name,value):
        if name is 'solutions':
            self.solver.solutions = value
        else:
            object.__setattr__(self,name,value)
        
    # ========================================================================
    #   Plotting methods
    # ========================================================================    
    def plotclaw(self, datadir='.'):
        pydir = '/home/rjl/claw/trunk/claw/python'
        if sys.platform in ['cygwin', 'win32']:
            syscmd = " C:/Python25/python.exe C:/cygwin%s/pyclaw/plotclaw.py %i %s" \
                   % (pydir, self.ndim, datadir) 
        else:
            syscmd = " python %s/pyclaw/plotclaw.py %i %s" \
                   % (pydir, self.ndim, datadir) 
        os.system(syscmd)
        
    def plot(self):
        """Plotting routine for """
        
        pass
    

    
    #------------------------------------------------------------
    def runxclaw(self, verbose=True):
    #------------------------------------------------------------
        '''
        Run the command xdir/xclawcmd, directing the output fort.*
        files to outdir, writing unit 6 timestepping info to file xclawout.
        Runtime error messages are written to file xclawerr.
        If xclawout(xclawerr) is None, then output to stdout(stderr).
    
        If savecode==True, archive a copy of the code into directory outdir.
    
        This function returns the returncode from the process running xclawcmd,
        which will be nonzero if a runtime error occurs.
        '''
        import os, shutil, glob
    
        try:
            import subprocess
        except:
            print 'Must use a more recent version of Python with module subprocess'
            print 'Python 2.5 is recommended'
            raise ClawError('subprocess module missing')
            return
    
        debug = False
        
	try:
	    # the following attributes are needed:
	    xdir = self.xdir
	    rundir = self.rundir
	    outdir = self.outdir
	    overwrite = self.overwrite
	    xclawcmd = self.xclawcmd
	    xclawout = self.xclawout
	    xclawerr = self.xclawerr
	    runmake = self.runmake
	    savecode = self.savecode
	except:
	    raise ClawError('missing attributes in runxclaw')
	    return

        startdir = os.getcwd()
        xdir = os.path.abspath(xdir)
        xclawcmd = os.path.join(xdir,xclawcmd)
    
        try:
            os.chdir(xdir)
        except:
            raise ClawError( "Cannot change to directory xdir = %s" %xdir)
            return 
    
        if runmake:
            try:
                os.system('make')
            except:
                print 'Warning: no make file in directory xdir = ',xdir
    

    
        if os.path.isfile(outdir):
            print "Error: outdir specified is a file"
            return
        elif (os.path.isdir(outdir) & overwrite):
            if verbose:
                print "Directory ", outdir, " already exists, "
                print "   fort.* files will be overwritten "
        elif (os.path.isdir(outdir) & (not overwrite)):
            print "Error in singleplot1: Directory ", outdir, " already exists"
            print "  Remove directory with 'rm -r ",outdir,"' and try again,"
            print "  or use overwrite=True in call to runxclaw"
            return
        else:
            try:
                os.mkdir(outdir)
            except:
                print "Cannot make directory ",outdir
                return
    
        try:
            os.chdir(outdir)
        except:
            print '*** Error in runxclaw: cannot move to outdir = ',\
                  outdir
            raise
            return
    
        fortfiles = glob.glob(os.path.join(outdir,'fort.*'))
        if overwrite:
            # remove any old versions:
            for file in fortfiles:
                os.remove(file)
        else:
            if len(fortfiles) > 1:
                print "*** Remove fort.* and try again,"
                print "  or use overwrite=True in call to runxclaw"
                return
            
        
        try:
            os.chdir(rundir)
        except:
            raise ClawError("Cannot change to directory %s" % rundir)
            return 
    
        datafiles = glob.glob('*.data')
        if datafiles == ():
            print "Warning: no data files found in directory ",rundir
        else:
            if rundir != outdir:
                for file in datafiles:
                    shutil.copy(file,os.path.join(outdir,file))
    
        if xclawout:
            xclawout = open(xclawout,'wb')
        if xclawerr:
            xclawerr = open(xclawerr,'wb')
    
        os.chdir(outdir)
    
        #print "\nIn directory outdir = ",outdir,"\n"
    
        # execute command to run fortran program:
    
        try:
            #print "\nExecuting ",xclawcmd, "  ...  "
            #pclaw = subprocess.Popen(xclawcmd,stdout=xclawout,stderr=xclawerr)
	    #print '+++ pclaw started'
            #pclaw.wait()   # wait for code to run
	    #returncode = pclaw.returncode
	    #print '+++ pclaw done'

	    returncode = os.system(xclawcmd)
    
            if returncode == 0:
                print "\nFinished executing\n"
            else:
                print "\n *** Runtime error: return code = %s\n " % returncode
        except:
            raise ClawError("Could not execute command %s" % xclawcmd)
	    return
    
        os.chdir(startdir)
        return returncode
