#!/usr/bin/env python
# encoding: utf-8
# ======================================================================
#  Package:     pyclaw
#  File:        claw_data.py
#  Created:     Aug 7, 2008
#  Author:      Kyle Mandli
# ============================================================================
"""
ClawData class file

Contains the general class definition and the subclasses ClawRunData and 
ClawFrameData.
"""
import shutil, os, copy, re, logging

from error import ClawError

# ============================================================================
#  General class definition for ClawData
# ============================================================================
class ClawData(object):
    """ClawData class
    
    Generalized class for Clawpack data.  Contains generic methods for reading
    and writing data to and from the object.
    
    Members:
        attributes = List of the names of the attributes of this class
        
    """
    
    __name__ = 'ClawData'
    
    attributes = []
    
    # ========================================================================
    #  Initialization routine
    # ========================================================================
    def __init__(self,data_files=[]):
        """Initialize a ClawData object
        
        Accepts a list of data_files to be read into and instantiate into one
        ClawData object.  An empty object can be created by not passing 
        anything in for the data_file_list
        """
        
        if isinstance(data_files, str):
            data_files = [data_files]
        
        self.__data = {}
        for attr in self.attributes:
            self.__data[attr] = {'value':None, 'required':True, 'owner':None}

        # Setup data logger
        self.logger = logging.getLogger('data')

        #print '<p> data_files = ',data_files;
        if len(data_files) > 0:
            #while isinstance(data_files[0],list) | isinstance(data_files[0],tuple):
            #    data_files = data_files[0]
            self.read(data_files)
        
    def setup(self,data_files=[]):
        """DEPRECATED  Sets up the ClawData object  DEPRECATED
        
        This method creates and sets up the internal data object and reads
        the initial data files passed in.  A subclass should call this method
        once it has created its attribute list.
        
        """
        self.__data = {}
        for attr in self.attributes:
            self.__data[attr] = {'value':None, 'required':True, 'owner':None}

        #print '<p> data_files = ',data_files;
        if len(data_files) > 0:
            #while isinstance(data_files[0],list) | isinstance(data_files[0],tuple):
            #    data_files = data_files[0]
            self.read(data_files)

    # ========================================================================
    # Return string representation of this ClawData Object
    # ========================================================================    
    def __str__(self):
        """String representation of this object"""
        output = "Required attributes:\n"
        for k in self.get_required():
            v = self.__data[k]['value']
            output += "%s = %s\n" % (str(k).rjust(25),str(v).ljust(12))
        output += "Parameters:\n"
        for k in self.get_parameters():
            if k is not 'logger':
                v = self.__data[k]['value']
                output += "%s = %s\n" % (str(k).rjust(25),str(v).ljust(12))
        output += "Owners:\n"
        for owner in self.get_owners():
            output += "%s\n" % str(owner).rjust(25)
        return output
    
    def data_repr(self):
        """Debug routine for checking the contents of __data."""
        output = "Attributes of data dictionary:\n"
        output += "Name".ljust(15)+"Value".rjust(25)+"Required".rjust(10) + \
            "  Owner".ljust(10) + "\n"
        for (k,v) in self.__data.iteritems():
            output += (k+":").ljust(15)
            output += "%s" % str(v['value']).rjust(25)
            output += "%s" % str(v['required']).rjust(10)
            output += "  %s" % str(v['owner']).ljust(10)
            output += "\n"
        return output

    # ========================================================================
    #  General access methods that are overridden
    # ========================================================================
    def __getattribute__(self,name):
        # print "Getting %s" % name
        if name is '__data' or name is '_ClawData__data':
            if issubclass(self.__class__,ClawData):
                return object.__getattribute__(self,'_ClawData__data')
            else:
                return object.__getattribute__(self,'__data')
        try:
            return object.__getattribute__(self,'_ClawData__data')[name]['value']   
        except(AttributeError,KeyError):
            return object.__getattribute__(self,name)
        except:
            raise
            
    def __setattr__(self,name,value):
        # print "Setting %s = %s" % (name,str(value))
        if name in ['_ClawData__data','attributes']:
            object.__setattr__(self,name,value)
        else:
            data = object.__getattribute__(self,'_ClawData__data')
            try:
                data[name]['value'] = value
            except KeyError:
                data[name] = {'value':value, 'required':False, 'owner':None}

    # ========================================================================
    #  Access methods to the data dictionary
    # ========================================================================
    def get_required(self):
        """Returns a list of the required data values."""
        return [k for (k,v) in self.__data.iteritems() if v['required']]
        
    def get_parameters(self):
        """Returns a list of the non-required data values."""
        return [k for (k,v) in self.__data.iteritems() if not v['required']]

    def get_owners(self,supplementary_file=None):
        """Returns a list of owners excluding the owner None.  If 
        supplementary_file is provided, None is replace by that owner."""
        owners = []
        for (key,values) in self.__data.iteritems():
            if values['owner'] is None:
                self.__data[key]['owner'] = supplementary_file
            # This simultaneously finds one instance of an owner and tests
            # to see if the supplementary_file is not None
            if values['owner'] not in owners and values['owner'] is not None:
                owners.append(values['owner']) 
        return owners
        
    def has_attribute(self,name):
        """Returns True if data object has the key name, else False."""
        return self.__data.has_key(name)
        
    def set_required(self,name,required):
        """Sets whether the name is required for the data object to be valid"""
        try:
            self.__data[name]['required'] = required
        except KeyError:
            raise
        except:
            raise
            
    def set_owner(self,name,owner):
        """Sets the owner of the given data."""
        try:
            self.__data[name]['owner'] = owner
        except KeyError:
            raise
        except:
            raise
            
    def remove_attributes(self, arg_list):
        """Remove the listed attributes."""
        
        # Convert to list if args is not already a list
        if not isinstance(arg_list,list):
            arg_list = [arg_list]
        
        for arg in arg_list:
            self.__data.pop(arg)
            self.attributes.remove(arg)
            
    # ========================================================================
    #  Read in a collection of data_files and create the corresponding fields
    #  and parameters
    # ========================================================================
    def read(self,data_paths):
        """ClawData.read(data_paths)
        
        Read data in from a clawpack style data file.
        Any lines of the form
           values  =:  name
        is used to set an attribute or parameter of self.
        
        Stores each value in a parameter dictionary unless it is a member of
        the attributes list.
        
        Arguments:
            data_paths = Path to a data file to be read in, can also be a list
                of files to be read in.
        """
        #print '<p> data_paths = ',data_paths
        if isinstance(data_paths,str):
            data_paths = [data_paths]
        for data_path in data_paths:
            try:
                data_path = os.path.abspath(data_path)
                data_file = open(data_path,'r')
            except IOError, msg:
                print msg
                raise

            regexp = re.compile(r"(?P<values>.*)=:(?P<name>.*)")
            try:
                for line in data_file:
                    result = regexp.search(line)
                    if result:
                        name = re.split(r'\s+', result.group('name').strip())[0]
                        values = re.split(r'\s+', result.group('values').strip())

                        if len(values) > 0:
                            if values[0] == 'T':
                                value = 'True'
                            elif values[0] == 'F':
                                value = 'False'
                            elif len(values) > 1:
                                value = '[' + ','.join(values) + ']'
                            else:
                                value = values[0]
                            required = name in self.attributes
                            exec_line = "self.__data[name] = " + \
                                "{'value':%s, 'required':required," % value +\
                                "'owner':data_path}"
                            exec(exec_line)
                        else:
                            print "Empty values for name " + name
                            #raise ClawError("Empty values for name " + name)
            except:
                print "Unable to read the following line from %s: \n%s" % (data_path,line)
                raise
            data_file.close()
                        
    # ========================================================================
    #  Write out the contents of this ClawData object
    # ========================================================================
    def write(self,data_files=None,supplementary_file=None):
        """ClawData.write(data_files=None,supplementary_file=None)
        
        This method writes out the current attributes and parameters of the 
        data object to a file or list of files.  The format for the output 
        will be in the form:
            values =: name
        The order is either retained from the files in the data list or
        written in the order they are in the list and dictionary iterators
        
        The behavior is determined by the arguments passed into the routine:
        No arguments:
            The contents of the object will be written out to the files listed
            in data_files only if each attribute or parameter is contained in
            the file.  This implies that if an attribute or parameter is not
            located in any of the files in data_files then it will not be
            written out to file.
        If data_files is provided and data_files is a valid owner:
            Write out the attributes appropriate to that file or files
        If data_files is provided and not(data_files in owner):
            Write all parameters and attributes to the data_file given
        If supplementary_file is provided
            Write all data attributes and parameter belonging to files into
            their respective places but use data_file to store unassigned 
            parameters and attributes.
        """        
        # Expand supplementary_file if it is not None
        if supplementary_file is not None:
            supplementary_file = os.path.abspath(supplementary_file)

        # Create list of owners
        owners = self.get_owners(supplementary_file=supplementary_file)
        
        #print 'in write: data_files = ',data_files
        # Write to the entire owner list
        if data_files is None:
            file_list = owners
        elif isinstance(data_files,str):
            #print '<p>in write'; import sys; sys.exit(0)
            path = os.path.abspath(data_files)
            if path not in owners:    
                # Create temporary data file to store all data in the object
                try:
                    temp_file = open(path,'w')
                    for key in self.__data.iterkeys():
                        temp_file.write("1 =: %s\n" % key)
                    temp_file.close()
                except IOError, (errno, strerror):
                    print "I/O error(%s): %s" % (errno, strerror)
                    raise
                except:
                    raise
            # Add the path to the file_list
            file_list = [path]
        # Check to make sure all the paths are in the owner list
        elif isinstance(data_files,list):
            for path in data_files:
                path = os.path.abspath(path)
                if not(path in owners):
                    print "%s is not a registered owner!"
                    return
            file_list = data_files
        else:
            raise ClawError("Invalid argument list given to write().")

        # Create temporary supplementary file if requested
        if supplementary_file is not None:
            try:
                sup_file = open(supplementary_file,'w')
                for (key,value) in self.__data.iteritems():
                    if value['owner'] == supplementary_file:
                        sup_file.write("-1 =: %s" % key)
                sup_file.close()
                file_list.append(supplementary_file)
            except:
                raise


        # Regular expression for searching each file
        regexp = re.compile(r"(?P<values>.*)=:(?P<name>.*)")
        # Loop over each file
        for data_path in file_list:
            # Open the data file and temporary file
            try:
                data_file = open(data_path,'r')
            except(IOError):
                raise
            try:
                temp_path = os.path.join(os.path.dirname(data_path), \
                    'temp.' + os.path.basename(data_path))
                temp_file = open(temp_path,'w')
            except(IOError):
                print "IOERROR"
                raise

            try:
                for line in data_file:
                    result = regexp.search(line)
                    if result:
                        name = re.split(r'\s+', result.group('name').strip())[0]
                        values = re.split(r'\s+', result.group('values').strip())

                        if len(values) == 0:
                            line = ''
                        elif self.__data[name]['owner'] == data_path or \
                                data_path not in owners:
                            newvalues = self.__data[name]['value']

                            # Convert newvalues to an appropriate string repr
                            if isinstance(newvalues,tuple) \
                                | isinstance(newvalues,list):
                                # Remove [], (), and ','
                                newstring = repr(newvalues)[1:-1]
                                newstring = newstring.replace(',','')
                            elif isinstance(newvalues,bool):
                                if newvalues:
                                    newstring = 'T'
                                else:
                                    newstring = 'F'
                            else:
                                newstring = repr(newvalues)

                            newstart = str.ljust(newstring,25)
                            line = line.replace(result.group('values') + "=:", \
                                newstart + " =:")
                        else:
                            print "Error writing out %s" % name
                            raise AttributeError, name

                    # Write the new line
                    temp_file.write(line)
            except:
                raise

            # Close files
            data_file.close()
            temp_file.close()

            # Rename the temporary file to the data_path name
            try:
                shutil.move(temp_path,data_path)
            except:
                raise
    
    def is_valid(self):
        """ClawData.is_valid() - Returns whether this objects attributes
                have all been set.

           returns True if this is a valid ClawData object, False otherwise
        """
        for attribute in self.attributes:
            exec('not_valid = self.%s is None' % attribute)
            if not_valid:
                self.logger.info('Invalid attribute %s' % attribute)
                return False
        return True

# ============================================================================
#  Subclass ClawGridData containing data for creating a ClawGrid object
# ============================================================================
class ClawGridData(ClawData):
    """ClawGridData class
    
    ClawData class for containing grid data
    
    Members:
        data_files = List of data_files from which this object was created
        attributes = List of the names of the attributes of this class
        ndim = Number of dimensions, note that this property must be set in
            order for the correct set of attributes to be defined.
        
        List of attributes:  These are expected to be set
            'ndim','mbc','meqn','maux','mcapa'
            
        Grid Dimension Attributes: These are set with the appropriate ndim 
                when appropriate.
            (mx,my,mz) = Number of cells in the associated direction
            (xlower,ylower,zlower) = Minimum coordinate in the dimension
            (xupper,yupper,zupper) = Maximum coordinate in the dimension
            (mthbc_xlower,mthbc_ylower,mthbc_zlower) = Lower boundary
                condition method to be used
            (mthbc_xupper,mthbc_yupper,mthbc_zupper) = Upper boundary
                condition method to be used
    """
    def __init__(self,data_file_list=[],ndim=None):
        """Initialize a ClawGridData object
        
        Accepts two different types of initialization:
          1) data_file_list = List of data files to be read in to instantiate
               a grid data object, not that this can be done by reading in a 
               valid data file for a ClawRunData object.
          2) ndim = Number of dimensions represented in this data object
               
        Note that ndim must be set from one of the data files, the run_data ''
        object, or directly via the key argument ndim.
        """
        
        if self.attributes is None:
            self.attributes = ['ndim','mbc','meqn','maux','mcapa','mx','my',
                'mz','xlower','ylower','zlower','xupper','yupper','zupper',
                'dx','dy','dz','mthbc_xlower','mthbc_ylower','mthbc_zlower',
                'mthbc_xupper','mthbc_yupper','mthbc_zupper']
        else:
            for attr in ['ndim','mbc','meqn','maux','mcapa','mx','my',
                    'mz','xlower','ylower','zlower','xupper','yupper',
                    'zupper','dx','dy','dz','mthbc_xlower','mthbc_ylower',
                    'mthbc_zlower','mthbc_xupper','mthbc_yupper',
                    'mthbc_zupper']:
                self.attributes.append(attr)
        
        # Initialize the data object and read the data files
        ClawData.__init__(self,data_file_list)       
        
        if self.ndim is None:
            if ndim is not None:
                self.ndim = ndim
            else:
                raise ValueError('Must set ndim')
            
        # Take away the attributes that are not needed due to ndim
        if self.ndim < 3:
            self.remove_attributes(['mz','zlower','zupper','mthbc_zlower',
                'mthbc_zupper','dz'])
            if self.ndim < 2:
                self.remove_attributes(['my','ylower','yupper','mthbc_ylower',
                    'mthbc_yupper','dy'])
    
    def construct_attributes(self):
        """Constructs (dx,dy,dz) or (xupper,yupper,zupper) if missing"""
        if self.dx is None:
            self.dx = float(self.xupper - self.xlower) / self.mx
        elif self.xupper is None:
            self.xupper = self.dx * self.mx + self.xlower
        if self.ndim > 1:
            if self.dy is None:
                self.dy = float(self.yupper - self.ylower) / self.my
            elif self.xupper is None:
                self.xupper = self.dy * self.my + self.ylower
            if self.ndim > 2:
                if self.dz is None:
                    self.dz = float(self.zupper - self.zlower) / self.mz
                elif self.zupper is None:
                    self.zupper = self.dz * self.mz + self.zlower 
                    
    def is_valid(self):
        """ClawGridData.is_valid() - Returns whether this objects attributes
                have all been set.
            
            Overriden from ClawData.is_valid() to check for amr parameters
            if set.

           returns True if this is a valid ClawRunData object, False otherwise
        """
    
        for attribute in self.attributes:
            # Treat dx,dy,dz and xupper,yupper,zupper attributes differently
            # since we need only one of the pair to procede
            if attribute in ['dx','dy','dz']:
                exec("not_valid = self.%s is None and self.%slower is None" %
                        (attribute,attribute[-1]))
                if not_valid:
                    self.logger.info('Invalid attribute is %s' % attribute)
                    return False
            elif attribute in ['xlower','ylower','zlower']:
                exec("not_valid = self.%s is None and self.d%s is None" %
                        (attribute,attribute[0]))
                if not_valid:
                    self.logger.info('Invalid attribute %s' % attribute)
                    return False
            else:
                exec('not_valid = self.%s is None' % attribute)
                if not_valid:
                    self.logger.info('Invalid attribute %s' % attribute)
                    return False
        return True
    
                    
# ============================================================================
#  Subclass ClawRunData containing data for running a simulation specifically
# ============================================================================
class ClawRunData(ClawGridData):
    """ClawRunData class
    
    ClawData class for containing run data.  
    
    This is a subclass of ClawGridData, please refer to it for additional
    attributes that are required to make this a valid data object.
    
    Members:
        data_files = List of data_files from which this object was created
        attributes = List of the names of the attributes of this class
        
        List of attributes:  These are expected to be set
            'nout','outstyle','tfinal','dt_initial',
            'dt_max','cfl_max','cfl_desired','max_steps','dt_variable',
            'order','order_trans','verbosity','src_split','mwaves',
            'mthlim','t0'
    """
    
    # ========================================================================
    #  Initialization routine
    # ========================================================================
    def __init__(self,data_file_list=[]):
        """Initialize a ClawRunData object
        
        Accepts a list of data_files to be read into and instantiate into one
        ClawRunData object.  An empty object can be created by not passing 
        anything in for the data_file_list
        
        Arugments:
            data_file_list = List of files to be read in to instantiate this
                object
            ndim = Number of dimensions
        """        
        # Append to the attribute list
        if self.attributes is None:
            self.attributes = ['nout','outstyle','tfinal','dt_initial',
                'dt_max','cfl_max','cfl_desired','max_steps','dt_variable',
                'order','order_trans','verbosity','src_split','mwaves',
                'mthlim','t0','amr_maxnest']
        else:
            for attr in ['nout','outstyle','tfinal','dt_initial','dt_max',
                    'cfl_max','cfl_desired','max_steps','dt_variable','order',
                    'order_trans','verbosity','src_split','mwaves','mthlim',
                    't0','amr_maxnest']:
                self.attributes.append(attr)

        # Initialize the data object and read the data files
        ClawGridData.__init__(self,data_file_list)
        
        # Change mthlim to be an array regardless of how long it is
        if not isinstance(self.mthlim,list) and self.mthlim is not None:
            self.mthlim = [self.mthlim]
            
        # This is a temporary setting as this has not been added to all data 
        # files yet
        self.amr_maxnest = 1

        # Check to see if amr_maxnest > 1
        if self.amr_maxnest > 1:
            # We need to initialize and set the amr parameters
            raise ClawError("AMR attributes not implemented yet!")

    # Internally, ClawRunData contains a ClawGridData object containing the 
    # appropriate data but accessing the data is transparent for the user

    # ========================================================================
    # Checks to make sure all of the attributes of the ClawRunData object are
    # set (not None).  This is overridden from the ClawData object to check 
    # for AMR attributes correctly
    # ========================================================================        
    def is_valid(self):
        """ClawRunData.is_valid() - Returns whether this objects attributes
                have all been set.
            
            Overriden from ClawData.is_valid() to check for amr parameters
            if set.

           returns True if this is a valid ClawRunData object, False otherwise
        """
        # Check to see if the superclass is valid
        if not ClawGridData.is_valid(self):
            return False
        # Otherwise check for the attributes we know about
        for attribute in ['nout','outstyle','tfinal','dt_initial',
                'dt_max','cfl_max','cfl_desired','max_steps','dt_variable',
                'order','order_trans','verbosity','src_split','mwaves',
                'mthlim','t0','amr_maxnest']:
            # Treat the amr attributes differently
            if attribute in ['amr_maxnest']:
                if self.amr_maxnest > 1:
                    self.logger.info('Invalid attribute %s' % attribute)
                    return False # We need to check amr parameters here
            else:
                exec('not_valid = self.%s is None' % attribute)
                if not_valid:
                    self.logger.info('Invalid attribute %s' % attribute)
                    return False
        return True

# ============================================================================
#  Subclass ClawPlotData containing data for plotting results
# ============================================================================
class ClawPlotData(ClawData):
    """ClawPlotData class
    
    ClawData subclass containing plot data.

    """
    # ========================================================================
    #  Initialization routine
    # ========================================================================
    def __init__(self, data_file_list=[], controller=None):
        """Initialize a ClawPlotData object
        
        Accepts a list of data_files to be read into and instantiate into one
        ClawPlotData object.  An empty object can be created by not passing 
        anything in for the data_file_list
        """
        self.attributes = ['rundir','plotdir','outdir','overwrite','plotter',\
                           'ion',\
                           'msgfile','printfig_format','afterframe',\
                           'html','movie','latex','latex_pdf','latex_fname', \
                           'html_homelink','printfigs', \
                           'print_framenos','print_fignos', \
			   'iplotclaw_fignos',\
                           'latex_framesperpage','latex_framesperline',\
			   'latex_figsperline', 'latex_pdf', \
			   '_mode',\
                           'clear_figs', 'setplot', 'setplot_caller', 'eagle',\
                           'plotitem_dict', 'params']

        # Initialize the data object and read the data files
        self.setup(data_file_list)

        # default values of attributes:

        if controller:
            controller.plotdata = self
            # inherit some values from controller
            self.rundir = copy.copy(controller.rundir)
            self.outdir = copy.copy(controller.outdir)
        else:
            self.rundir = os.getcwd()     # uses *.data from rundir
            self.outdir = os.getcwd()     # where to find fort.* files


        self.plotdir = os.getcwd()      # directory for plots *.png, *.html
        self.overwrite = True           # ok to overwrite old plotdir?
        self.plotter = 'matplotlib'     # backend for plots
        self.msgfile = ''               # where to write error messages

        self.ion = False                # call ion() or ioff()?

        self.printfigs = True 
        self.print_format = 'png'     
        self.print_framenos = 'all'  # which frames to plot
        self.print_fignos = 'all'    # which figures to plot each frame

        self.iplotclaw_fignos = 'all'    # which figures to plot interactively

        self.latex = True                # make latex files for figures
        self.latex_fname = 'plots'       # name of latex file
        self.latex_framesperpage = 'all' # number of frames on each page
        self.latex_framesperline = 2     # number of frames on each line
        self.latex_figsperline = 'all'   # number of figures on each line
        self.latex_pdf = False           # run pdflatex on latex file

        self.html = True                # make html files for figures
        self.html_homelink = None       # link to here from top of _PlotIndex.html
        self.html_movie = True          # make html with java script for movie
        self.html_eagle = False         # use EagleClaw titles on html pages?

        self.gif_movie = False          # make animated gif movie of frames


    #    self.clear_figs = True          # give clf() command in each figure
                                        # before plotting each frame

        self.setplot = False            # Execute setplot.py in plot routine
    #    self.setplot_caller = None      # Set before calling setplot

        self.afterframe = None          # function called after all plots 
                                        # in each frame are done

        self.plotfigure_dict = {}  

        self.framesoln_dict = {}        # dictionary for holding framesoln
                                        # objects associated with plots

        self.refresh_frames = False     # don't re-read framesoln if 
                                        # already in framesoln_dict


        self._fignames = []
        self._fignos = []
        self._mode = 'unknown'

        #if data_file_list is not None:
        if len(data_file_list) > 0:
            # values in data files may overwrite some default values
            # or set parameter values in params dictionary
            for data_file in data_file_list:
                self.read(data_file)

    def new_plotfigure(self, name, figno):
        """
        Create a new figure that will be plotted for each frame.
        """
        if (self._mode != 'iplotclaw') and (name in self._fignames):
            print '*** Warning, figure named %s has already been created' % name
        if (self._mode != 'iplotclaw') and (figno in self._fignos):
            print '*** Warning, figure number %s has already been created' % figno
        if name not in self._fignames:
            self._fignames.append(name)
            self._fignos.append(figno)
        plotfigure = ClawPlotFigure(name, figno, self)
        self.plotfigure_dict[name] = plotfigure
        return plotfigure


    def getframe(self,frameno,outdir=None):
        """
        Return an object of class ClawSolution containing the solution
        for frame number frameno.

        If self.refresh_frames == True then this frame is read from the fort
        files, otherwise it is read from the fort files only if the
        the dictionary self.framesoln_dict has no key frameno.  If it does, the
        frame has previously been read and the dictionary value is returned.
        """

        import solution

	framesoln_dict = self.framesoln_dict
	if outdir:
	    key = (frameno, outdir)
            outdir = os.path.abspath(outdir)
	else:
	    key = frameno
	    outdir = self.outdir

        if not framesoln_dict.has_key(key):
            thisdir = os.getcwd()
            try:
                os.chdir(outdir)
            except:
                print '*** Error in getframe: cannot move to outdir = ',\
                       outdir
                print '*** thisdir = ',thisdir
                raise
                return
	    try:
                framesoln = solution.ClawSolution(frameno)
	    except:
	        print '*** error reading frame'
            os.chdir(thisdir)
            framesoln_dict[key] = framesoln
            print '--> ClawPlotData.getframe: read frame %s at t = %g' \
                    % (frameno,framesoln.t)
	    if key != frameno:
	        print '    from outdir = ',outdir
        else:
            framesoln = self.framesoln_dict[key]

        return framesoln


    def clearframes(self, framenos='all'):
        """
        Clear one or more frames from self.framesoln_dict.
        Need to add outdir option!
        """

        if isinstance(framenos, int):
            framenos = [framenos]  # turn into a list

	if framenos=='all':
            self.framesoln_dict.clear()
            print 'Cleared all frames'
	else:
            for frameno in framenos:
                xxx = self.plotdata.framesoln_dict.pop(frameno,None)
                if xxx is None:
                   print 'No frame data to clear for frame ',frameno
                else:
                   print 'Cleared data for frame ',frameno


    def plotframe(self, frameno):
        import plotting
        plotting.plotframe(frameno, self)
        
    def printframes(self):
        import plotting
        plotting.printframes(self)
        
    def fignos(self):
        """
        Return a list of the figure numbers actually used.
        Useful in afterframe function for example to loop over all
        figures and do something.
        """
        return self._fignos

    def mode(self):
        """
        Return self._mode, which is set internally to 
           'iplotclaw' if Iplotclaw is in use,
           'printframes' if printframes is being used
        Useful in afterframe function if you want to do different things
           for interactive or print modes.
        """
        return self._mode

    def iplotclaw(self):
        """
        Return True if interactive plotting with iplotclaw is being done.
        """
        return (self._mode == 'iplotclaw')


    def getfigure(self,figname):
        try:
            plotfigure = self.plotfigure_dict[figname]
        except:
            print '*** Error accessing plotfigure_dict[%s]' % figname
            return None
        return plotfigure

    def getaxes(self,axesname,figname=None):
        found = True
        if not figname:
            found = False
            for fig in self._fignames:
                plotfigure = self.getfigure(fig)
                if axesname in plotfigure._axesnames:
                    if found == True: # already found!
                        print '*** Ambiguous... must specify figname'
                        print '    try getaxes(axesname, figname)'
                        return None
                    figname = fig
                    found = True
        if not found:
            print '*** No axes found with name = ',axesname
            return None
        try:
            plotfigure = self.getfigure(figname)
            plotaxes = plotfigure.plotaxes_dict[axesname]
        except:
            print '*** Error accessing plotaxes[%s]' % axesname
            print '*** figname = %s' % figname
            return None
        return plotaxes

    def getitem(self,itemname,axesname=None,figname=None):
        found = True
        if not figname:
            # search over all figures looking for the item
            found = False
            for fign in self._fignames:
                plotfigure = self.getfigure(fign)
                if not axesname:
                    # search over all axes looking for the item
                    for axesn in plotfigure._axesnames:
                        plotaxes = self.getaxes(axesn,fign)
                        if itemname in plotaxes._itemnames:
                            if found == True: # already found!
                                print '*** Ambiguous... must specify figname and/or axesname'
                                print '    try getitem(itemname, axesname, figname)'
                                return None
                            axesname = axesn
                            figname = fign
                            found = True
                else:
                    # axesname was specified (but not figname)
                    plotaxes = self.getaxes(axesname,fign)
                    if itemname in plotaxes._itemnames:
                        if found == True: # already found!
                            print '*** Ambiguous... must specify figname and/or axesname'
                            print '    try getitem(itemname, axesname, figname)'
                            return None
                        figname = fign
                        found = True

        elif not axesname:
            # figname was specified but not axesname.
            # search over all axes looking for the item
            found = False
            plotfigure = self.getfigure(figname)
            for axesn in plotfigure._axesnames:
                plotaxes = self.getaxes(axesn,figname)
                if itemname in plotaxes._itemnames:
                    if found == True: # already found!
                        print '*** Ambiguous... must specify axesname'
			print '    try getitem(itemname, axesname, figname)'
                        return None
                    axesname = axesn
                    found = True

        if not found:
            print '*** No item found with name = ',itemname
            return None
        try:
            plotaxes = self.getaxes(axesname,figname)
            plotitem = plotaxes.plotitem_dict[itemname]
        except:
            print '*** Error accessing plotitem[%s]' % itemname
            print '*** figname = ',figname
            print '*** axesname = ',axesname
            return None
        return plotitem


    def showitems(self):
        fignames = self._fignames
        print "\n\nCurrent plot figures, axes, and items:"
        print "---------------------------------------"
        for figname in fignames:
            plotfigure = self.getfigure(figname)
            s =  "  figname = %s, figno = %s" % (figname, plotfigure.figno)
            if not plotfigure._show: 
                s = s + "  [Not showing]"
            print s
            axesnames = plotfigure._axesnames
            for axesname in axesnames:
                plotaxes = self.getaxes(axesname,figname)
                s =  "     axesname = %s, axescmd = %s" \
                       % (axesname, plotaxes.axescmd)
                if not plotaxes._show: 
                    s = s + "  [Not showing]"
                print s
                for itemname in plotaxes._itemnames:
                    plotitem = self.getitem(itemname,axesname,figname)
                    plot_type = plotitem.plot_type
                    s =  "        itemname = %s,  plot_type = %s" \
                          % (itemname,plot_type)
                    if not plotitem._show: 
                        s = s + "  [Not showing]"
                    print s
            print " "


    def getq(self,frameno):
        solution = self.getframe(frameno)
        grids = solution.grids
        if len(grids) > 1:
            print '*** Warning: more than 1 grid, q on grid[0] is returned'
        q = grids[0].q
        return q


# ============================================================================
#  Subclass ClawPlotFigure containing data for plotting a figure
# ============================================================================
class ClawPlotFigure(ClawData):
    """
    
    ClawData subclass containing plot data needed to plot a single figure.
    This may consist of several ClawPlotAxes objects.

    """

    # ========================================================================
    #  Initialization routine
    # ========================================================================
    def __init__(self, name, figno, plotdata):
        """
        Initialize a ClawPlotFigure object
        """

        self.attributes = ['name','figno','_plotdata','clf','plotaxes_dict', \
                           '_axesnames','show','_show','kwargs','_handle']

        self.setup()
        self._plotdata = plotdata           # parent ClawPlotData object
        #self.name = name
        self.figno = figno
        self.kwargs = {}
        self.clf_each_frame = True
        self._axesnames = []
        self.show = True
        self._show = True
        self.plotaxes_dict  = {}

    def new_plotaxes(self, name):
        """
        Create a new axes that will be plotted in this figure.
        """
        if name in self._axesnames:
            print '*** Warning, axes named %s has already been created' % name

        if name not in self._axesnames:
            self._axesnames.append(name)
        plotaxes = ClawPlotAxes(name, self)
        self.plotaxes_dict[name] = plotaxes
        return plotaxes

    def gethandle(self):
        _handle = getattr(self,'_handle',None)
        return _handle


# ============================================================================
#  Subclass ClawPlotAxes containing data for plotting axes within a figure
# ============================================================================
class ClawPlotAxes(ClawData):
    """
    
    ClawData subclass containing plot data needed to plot a single axes.
    This may consist of several ClawPlotItem objects.

    """

    # ========================================================================
    #  Initialization routine
    # ========================================================================
    def __init__(self, name, plotfigure):
        """
        Initialize a ClawPlotAxes object
        """

        self.attributes = ['name','figno','plotdata','plotfigure','title',\
                           'axescmd','xlim','ylim','plotitem_dict', \
                           'afteraxes','_itemnames','show','_show','_handle', \
                           '_plotfigure','_plotdata']
        self.setup()

        self._plotfigure = plotfigure                   # figure this item is on
        self._plotdata = plotfigure._plotdata           # parent ClawPlotData object

        self.name = name
        self.title = name
        self.title_with_t = True        # creates title of form 'title at time t = ...'
        self.axescmd = 'subplot(1,1,1)'
        self.afteraxes = None
        self.xlim = None
        self.ylim = None
        self.plotitem_dict  = {}
        self._itemnames = []
        self.show = True
        self._show = True
        self._handle = None

    def new_plotitem(self, name, plot_type):
        # Create a new entry in self.plotitem_dict

        if name not in self._itemnames:
            self._itemnames.append(name)

        plotitem = ClawPlotItem(name, plot_type, plotaxes=self)

        self.plotitem_dict[name] = plotitem
        return plotitem

    def gethandle(self):
        _handle = getattr(self,'_handle',None)
        return _handle

# ============================================================================
#  Subclass ClawPlotItem containing data for plotting a single object
# ============================================================================
class ClawPlotItem(ClawData):
    """
    
    ClawData subclass containing plot data needed to plot a single object.
    This may be a single curve, set of points, contour plot, etc.

    """

    # ========================================================================
    #  Initialization routine
    # ========================================================================
    def __init__(self, name, plot_type, plotaxes):
        """
        Initialize a ClawPlotItem object
        """
        self.attributes = ['ndim','outdir',
                           'plot_type','plot_var',\
                           'figno',  'params', \
                           '_show','aftergrid','framesoln_dict', \
                           '_plotaxes','_plotfigure','_plotdata']  

        self.setup()


        self._plotaxes = plotaxes                       # axes this item is on
        self._plotfigure = plotaxes._plotfigure         # figure this item is on
        self._plotdata = plotaxes._plotfigure._plotdata           # parent ClawPlotData object

        try:
            ndim = int(plot_type[0])   # first character of plot_type should be ndim
        except:
            print '*** Error: could not determine ndim from plot_type = ',plot_type

        self.ndim = ndim
        self.name = name
        self.outdir = None              # indicates data comes from 
                                        #   self._plotdata.outdir

        self.plot_type = plot_type
        self.plot_var = 0

        self.aftergrid = None           # function called after each grid is
                                        # plotted within each single plot.

        self.show = True                # False => suppress showing this item
        self._show = True                # Internal


        if ndim == 1:
            self.attributes += [ 'plotstyle', 'color', 'kwargs']

            # default values specifying this single plot:
            self.plotstyle = '-'
            self.color = 'b'
            self.kwargs = {}

            if plot_type == '1d_from_2d_data':
                self.map2d_to_1d = None   # mapping function
                self.attributes += [ 'map2d_to_1d']

        elif ndim == 2:

            # default values specifying this single plot:
            self.plot_type = plot_type
            self.plot_grid = 0
            self.grid_color = 'k'
            self.plot_grid_edges = 0
            self.grid_edge_color = 'k'

            if plot_type == '2d_pcolor':
                from pylab import cm
                self.pcolor_cmap = cm.RdYlBu
                self.pcolor_cmin = None
                self.pcolor_cmax = None
                self.pcolor_colorbar = True
                self.attributes = self.attributes + \
                     ['pcolor_cmap', 'pcolor_cmin', 'pcolor_cmax', \
                      'pcolor_colorbar']
            elif plot_type == '2d_contour':
                self.contour_nlevels = 20
                self.contour_levels = None
                self.contour_min = None
                self.contour_max = None
                self.contour_kwargs = {'colors': 'k'}
                self.attributes = self.attributes + \
                     ['contour_nlevels', 'contour_levels', 'contour_min', \
                      'contour_max', 'contour_kwargs']

            elif plot_type == '2d_schlieren':
                 print '*** Warning: 2d_schlieren not yet implemented'

            else:
                 print '*** Warning 2d plot type %s not recognized' % plot_type

        else:
            self.attributes = []  
            print '*** Warning- ClawPlotItem not yet set up for ndim =',\
                  ndim

        self.params = {}  # dictionary to hold optional parameters


    def getframe(self,frameno):
        """
        Return an object of class ClawSolution containing the solution
        for frame number frameno.

        If self.refresh_frames == True then this frame is read from the fort
        files, otherwise it is read from the fort files only if the
        the dictionary self.framesoln_dict has key frameno.  If it does, the
        frame has previously been read and the dictionary value is returned.
        """

        plotdata = self._plotdata
        outdir = self.outdir
        framesoln = plotdata.getframe(frameno, outdir)

        return framesoln
