#!/usr/bin/env python
# encoding: utf-8
#  ======================================================================
#  Package:     pyclaw
#  File:        grid.py
#  Created:     Feb 6, 2008
#  Author:      Kyle Mandli
#  ======================================================================
"""Pyclaw grid routines

Module containing all gridding routines.  Not prepared for AMR as of yet!

"""

import os
import copy
import logging

from numpy.core.multiarray import empty, concatenate, zeros
from numpy.core.numeric import asarray

from data import ClawGridData
from error import ClawError

class ClawGrid(object):
    """Basic grid class for pyclaw
    
    Basic grid class containing information about the grid and the state
    vector q.

    Members:
        Solution -
            solution = Solution object that this grid is apart of
    
        Grid Information -
            level = Level of current grid in AMR structure
            gridno = Grid number of current grid
            ndim = Number of dimensions
            t = Current time on this grid
            mbc = Number of ghost cells along the boundaries
            meqn = Dimension of q array for this grid
            maux = Dimension of auxillary array for this grid
            mcapa = Auxillary array entry representing the capacity of the
              grid, this must be 0 if maux is 0
            (mx,my,mz) = Number of cells in the associated direction
            (xlower,ylower,zlower) = Minimum coordinate in the dimension
            (xupper,yupper,zupper) = Maximum coordinate in the dimension
            (dx,dy,dz) = Computational grid cell size
            (mthbc_xlower,mthbc_ylower,mthbc_zlower) = Lower boundary
                condition method to be used
            (mthbc_xupper,mthbc_yupper,mthbc_zupper) = Upper boundary
                condition method to be used
        
        Data Arrays - 
            q = Data for this grid of dimension (mx,my,mz,meqn)
            aux = Auxillary array for this grid of (*,*,*,maux), this array
              can be larger than the grid size to accomodate ghost cell values
              and other possibilities.
        
        Grid Arrays - 
                center[0]  center[1]        center[mx-2]  center[mx-1]
         edge[0]  |  edge[1]  |       edge[mx-2] | edge[mx-1]|  edge[mx]
            V     V     V     V            V     V     V     V     V
            | ----+---- | ----+---- | .... | ----+---- | ----+---- | 
            (xedge,yedge,zedge) = Grid of size (mx+1,my+1,mz+1) with location
                of left edge of each cell (including 1 ghost cell on right)
            (xcenter,ycenter,zcenter) = Grid of size (mx,my,mz) with location
                of center of each cell
    """
        
# ----------------------------------------------------------------------
    def __init__(self,data=None,ndim=None):
        """Create ClawGrid object from ClawData

        Creates a ClawGrid object based on the arguments provided.  Valid 
        arguments include the following:
            data = Valid ClawRunData or ClawGridData object
            ndim = Dimension of grid requested
        Note that if neiher a ClawGridData object or ndim is passed, the grid
        object is initialized with a dimension of 1.
            
        """
        # Viewable members of the class
        self.viewable_members = ['level','gridno','t','mbc','mcapa','meqn',
                                    'ndim','maux']
        
        # Set ndim, this needs to be done first in order to set the correct
        # default attribute list
        self.ndim = None
        if data is not None:
            if data.is_valid():
                self.ndim = data.ndim
        if self.ndim is None:
            if ndim is not None:
                self.ndim = ndim
            else:
                self.ndim = 1
        
        # Empty data values
        self.t = None
        self.mbc = None
        self.maux = None
        self.mcapa = None
        self.meqn = None
        for attr in ['mx','xlower','xupper','dx','mthbc_xlower',\
                'mthbc_xupper']:
            exec("self.%s = None" % attr)
            self.viewable_members.append(attr)
        if self.ndim > 1:
            for attr in ['my','ylower','yupper','dy','mthbc_ylower',\
                    'mthbc_yupper']:
                exec("self.%s = None" % attr)
                self.viewable_members.append(attr)
            if self.ndim > 2:
                for attr in ['mz','zlower','zupper','dz','mthbc_zlower',\
                        'mthbc_zupper']:
                    exec("self.%s = None" % attr)
                    self.viewable_members.append(attr)
                    
        # AMR Label for this grid, should be set later if needed
        self.level = 1
        self.gridno = 1
        
        # Set values via data object
        if data is not None:
            if data.is_valid():
                for attr in data.attributes:
                    exec('self.%s = data.%s' % (attr,attr))
                
        # q and aux array
        self.q = None
        self.aux = None
        
        # Mapping and boundary condition functions
        self.mapc2p = self.mapc2p_default
        if self.ndim == 1:
            self.qbc = self.qbc1_default
        else: 
            # There are no default routines for the boundary conditions
            # in python yet for ndim > 1
            self.qbc = None
        
        # Attempt to construct the grid arrays
        try:
            self.construct()
        except:
            pass
        
    def construct(self):
        """ClawGrid.constuct_grid - Constructs the arrays in this grid object
        
        This routine uses the values set in (mx,my,mz), (xlower,ylower,zlower)
        and (xupper,yupper,zupper) to initialize the arrays *edge and *center.
        
        """
        # Construct missing attributes if need be
        self.construct_attributes()

        # Construct grid arrays
        self.construct_edge_arrays()
        self.construct_center_arrays()   

        # Should setaux and qinit be called here, I do not think so...

    def construct_attributes(self):
        """Constructs (dx,dy,dz) or (xupper,yupper,zupper) if missing"""
        if self.dx is None:
            self.dx = float(self.xupper - self.xlower) / self.mx
        elif self.xupper is None:
            self.xupper = self.dx * self.mx + self.xlower
        if self.ndim > 1:
            if self.dy is None:
                self.dy = float(self.yupper - self.ylower) / self.my
            elif self.yupper is None:
                self.yupper = self.dy * self.my + self.ylower
            if self.ndim > 2:
                if self.dz is None:
                    self.dz = float(self.zupper - self.zlower) / self.mz
                elif self.zupper is None:
                    self.zupper = self.dz * self.mz + self.zlower 

    def construct_edge_arrays(self):
        """Constructs the arrays (xedge,yedge,zedge)"""
    
        self.xedge = empty(self.mx+1)
        for i in xrange(0,self.mx):
            self.xedge[i] = self.xlower + i*self.dx
        self.xedge[self.mx] = self.xlower + self.mx * self.dx
        if self.ndim >= 2:
            self.yedge = empty(self.my+1)
            for i in xrange(0,self.my):
                self.yedge[i] = self.ylower + i*self.dy
            self.yedge[self.my] = self.ylower + self.my * self.dy # Right edge    
            if self.ndim >= 3:
                self.zedge = empty(self.mz+1)
                for i in xrange(0,self.mz):
                    self.zedge[i] = self.zlower+i*self.dz
                self.zedge[self.mz] = self.zlower + self.mz*self.dz # Right edge
            
    def construct_center_arrays(self):
        """Constructs the arrays (xcenter,ycenter,zcenter)"""
 
        # Construct xcenter
        self.xcenter = empty(self.mx)
        for i in xrange(0,self.mx):
            self.xcenter[i] = self.xlower + (i+0.5)*self.dx
        if self.ndim >= 2:
            # Construct ycenter
            self.ycenter = empty(self.my)
            for i in xrange(0,self.my):
                self.ycenter[i] = self.ylower + (i+0.5) * self.dy
            if self.ndim >= 3:
                # Construct zcenter
                self.zcenter = empty(self.mz)
                for i in xrange(0,self.mz):
                    self.zcenter[i] = self.zlower + (i+0.5) * self.dz
        
    # ========================================================================
    #  Return string representation
    # ========================================================================
    def __str__(self):
        """String representation of this object"""
        output = "Required attributes:\n"
        for k in self.viewable_members:
            exec('v = self.%s' % k)
            output += "%s = %s\n" % (str(k).rjust(25),str(v).ljust(12))
        return output
    
    def is_valid(self):
        """Checks to see if all required attributes are present for a solver.
        
        Returns:  True if the grid is valid, False otherwise
        
        """
        valid_list = ['level','gridno','t','mbc','mcapa','meqn','ndim','maux',
                      'q']
        if self.maux > 0:
            valid_list.append('aux')
        if self.ndim > 0:
            valid_list.extend(['mx','xlower','xupper','dx','mthbc_xlower',\
                'mthbc_xupper'])
            if self.ndim > 1:
                valid_list.extend(['my','ylower','yupper','dy','mthbc_ylower',\
                    'mthbc_yupper'])
                if self.ndim > 2:
                    valid_list.extend(['mz','zlower','zupper','dz','mthbc_zlower',\
                        'mthbc_zupper'])
        for attr in valid_list:
            exec("valid = self.%s is not None" % attr)
            if not valid:
                logging.critical("Attribute %s was not valid!" % attr)
                return False
        return True
    
    def grid_data(self):
        """Returns a ClawGridData object representing the data from this grid
        
        This returns a reconstructed data object with deepcopied attributes
        based on a ClawGridData object.
        
        returns an object of type ClawGridData
        """
        
        data = ClawGridData(ndim=self.ndim)
        for attr in data.attributes:
            exec('data.%s = copy.deepcopy(self.%s)' % (attr,attr))
        
        return data
        
    #  =======================================================================
    #  Copy functions
    #  =======================================================================        
    def __copy__(self):
        return self.__class__(self)
    
    def __deepcopy__(self, memo={}):
        # Create basic container
        result = self.__class__()
        result.__init__(self.grid_data())
        
        # Copy time, level and gridno
        result.level = copy.deepcopy(self.level)
        result.gridno = copy.deepcopy(self.gridno)
        result.t = copy.deepcopy(self.t)
        
        # Copy over the q and aux arrays
        result.q = copy.deepcopy(self.q)
        if self.aux is not None:
            result.aux = copy.deepcopy(self.aux)
        
        return result
        
    # ========================================================================
    #  Map from computational space to physical space
    # ========================================================================
    def mapc2p_default(self,*args):
        """Default version of mapc2p, just returns the coordinates provided"""
        if len(args) == 1:
            return args[0]
        elif len(args) == 2:
            return args[0],args[1]
        elif len(args) == 3:
            return args[0],args[1],args[2]
        else:
            raise ClawError("Invalid number of dimensions")
        return None

    # ========================================================================
    #  Default boundary condition routines
    # ========================================================================
    def qbc1_default(self):
        """Default boundary conditions for 1D

        Returns:  qbc
            qbc = q vector with boundary ghost cells added
        """
        
        # Add ghost cells to q
        qbc = concatenate( (zeros((self.mbc,self.meqn)), \
                                self.q, zeros((self.mbc,self.meqn))) )
        
        # Left boundary
        if self.mthbc_xlower == 0:
            raise ClawError('User defined left BC requested but not ' + \
                'defined.  Specify a function handle to use custom ' +\
                'boundary conditions')
        elif self.mthbc_xlower == 1:
            # Zero-order extrapolation
            for i in xrange(self.mbc):
                qbc[i,:] = self.q[0,:]
        elif self.mthbc_xlower == 2:
            # Periodic
            qbc[:self.mbc,:] = self.q[-self.mbc:,:]
        elif self.mthbc_xlower == 3:
            # Solid wall, assumes second component is velocity or x momentum
            qbc[:self.mbc,:] = flipud(self.q[:self.mbc,:])
            qbc[:self.mbc,2] = -qbc[:self.mbc,2]
        else:
            raise ClawError('Left boundary condition option not implemented!')
            
        # Right boundary
        if self.mthbc_xupper == 0:
            raise ClawError('User defined right BC requested but not ' + \
                'defined.  Specify a function handle to use custom ' +\
                'boundary conditions')
        elif self.mthbc_xupper == 1:
            # Zero-order extrapolation
            for i in xrange(self.mbc):
                qbc[-i-1:,:] = self.q[-1,:]
        elif self.mthbc_xupper == 2:
            # Periodic
            qbc[-self.mbc:,:] = self.q[:self.mbc,:]
        elif self.mthbc_xupper == 3:
            # Solid wall, assumes second component is velocity or x momentum
            qbc[-self.mbc:,:] = flipud(self.q[-self.mbc:,:])
            qbc[-self.mbc:,2] = -qbc[-self.mbc:,2]
        else:
            raise ClawError('Right boundary condition option not implemented!')
        
        return qbc
        
    def qbc2_default(self):
        """Default boundary conditions for 2D

        Returns:  qbc
            qbc = q vector with boundary ghost cells added
        """
        raise ClawError("2D default qbc has not been implemented in python.")
        
    def qbc3_default(self):
        """Default boundary conditions for 1D

        Returns:  qbc
            qbc = q vector with boundary ghost cells added
        """
        raise ClawError("3D default qbc has not been implemented in python.")
# ----------------------------------------------------------------------

