#!/usr/bin/env python
# encoding: utf-8
#  ======================================================================
#  Package:     pyclaw.io
#  File:        ascii.py
#  Created:     Feb 7, 2008
#  Author:      Kyle Mandli
#  ======================================================================
"""Routines for reading and writing an ascii output file"""

import os,sys
import logging

from numpy.core.multiarray import empty
from pyclaw.grid import ClawGrid
from pyclaw.util import read_data_line

logger = logging.getLogger('io')

def write_ascii(solution,frame,path,file_prefix='fort',write_aux=False,
                    options={}):
    """Write out ascii data file
    
    Write out an ascii file formatted identical to the fortran clawpack files
    including writing out fort.t, fort.q, and fort.aux if necessary.  Note
    that there are some parameters that assumed to be the same for every grid
    in this format which is not necessarily true for the actual data objects.
    Make sure that if you use this output format that all of you grids share
    the appropriate values of ndim, meqn, maux, and t.
    
    Arguments:
      solution = Pyclaw object to be output
      frame = Frame number
      path = Root path
      file_prefix = Prefix for the file name, defaults to standard 'fort'
      write_aux = Boolean controlling whether the associated auxillary array 
        should be written out.
      options = Optional argument dictionary which can contain the following
        keyword/value pairs:
            EMPTY
    
    """
    
    # Option parsing
    option_defaults = {}
    for (k,v) in option_defaults.iteritems():
        if options.has_key(k):
            exec("%s = options['%s']" % (k,k))
        else:
            exec('%s = v' % k)
    
    try:
        # Create file name
        file_name = '%s.t%s' % (file_prefix,str(frame).zfill(4))
        f = open(os.path.join(path,file_name),'w')
        
        # Header for fort.txxxx file
        f.write("%18.8e     time\n" % solution.t)
        f.write("%5i                  meqn\n" % solution.meqn)
        f.write("%5i                  ngrids\n" % len(solution.grids))
        f.write("%5i                  ndim\n" % solution.ndim)
        f.write("%5i                  maux\n" % solution.maux)
        f.close()
        
        # Open fort.qxxxx for writing
        file_name = 'fort.q%s' % str(frame).zfill(4)
        q_file = open(os.path.join(path,file_name),'w')
        
        # If maux != 0 then we open up a file to write it out as well
        if solution.maux > 0 and write_aux:
            file_name = 'fort.a%s' % str(frame).zfill(4)
            aux_file = open(os.path.join(path,file_name),'w')
        
        # for i in range(0,len(solution.grids)):
        for grid in solution.grids:
            # Header for fort.qxxxx file
            q_file.write("%5i                  grid_number\n" % grid.gridno)
            q_file.write("%5i                  AMR_level\n" % grid.level)
            q_file.write("%5i                  mx\n" % grid.mx)
            if grid.ndim > 1:
                q_file.write("%5i                  my\n" % grid.my)
            if grid.ndim > 2:
                q_file.write("%5i                  mz\n" % grid.mz)
            q_file.write("%18.8e     xlow\n" % grid.xlower)
            if grid.ndim > 1:
                q_file.write("%18.8e     ylow\n" % grid.ylower)
            if grid.ndim > 2:
                q_file.write("%18.8e     zlow\n" % grid.zlower)
            q_file.write("%18.8e     dx\n" % grid.dx)
            if grid.ndim > 1:
                q_file.write("%18.8e     dy\n" % grid.dy)
            if grid.ndim > 2:
                q_file.write("%18.8e     dz\n" % grid.dz)
            q_file.write("\n")
            
            # Write data from q
            q = grid.q
            if grid.ndim == 1:
                for k in range(grid.mx):
                    for m in range(solution.meqn):
                        q_file.write("%16.8e" % q[k,m])
                    q_file.write('\n')
            elif grid.ndim == 2:
                for j in range(grid.my):
                    for k in range(grid.mx):
                        for m in range(solution.meqn):
                            q_file.write("%16.8e" % q[k,j,m])
                        q_file.write('\n')
                    q_file.write('\n')
            elif grid.ndim == 3:
                for l in range(grid.mz):
                    for j in range(grid.my):
                        for k in range(grid.mx):
                            for m in range(solution.meqn):
                                q_file.write("%16.8e" % q[k,j,l,m])
                            q_file.write('\n')
                    q_file.write('\n')
                q_file.write('\n')
            else:
                raise ClawError("Dimension error in writing fort file.")
            
            
            if solution.maux > 0 and write_aux:
                aux = grid.aux
                
                aux_file.write("%5i                  grid_number\n" % grid.gridno)
                aux_file.write("%5i                  AMR_level\n" % grid.level)
                aux_file.write("%5i                  mx\n" % grid.mx)
                if grid.ndim > 1:
                    aux_file.write("%5i                  my\n" % grid.my)
                if grid.ndim > 2:
                    aux_file.write("%5i                  mz\n" % grid.mz)
                aux_file.write("%18.8e     xlow\n" % grid.xlower)
                if grid.ndim > 1:
                    aux_file.write("%18.8e     ylower\n" % grid.ylower)
                if grid.ndim > 2:
                    aux_file.write("%18.8e     zlower\n" % grid.zlower)
                aux_file.write("%18.8e     dx\n" % grid.dx)
                if grid.ndim > 1:
                    aux_file.write("%18.8e     dy\n" % grid.dy)
                if grid.ndim > 2:
                    aux_file.write("%18.8e     dz\n" % grid.dz)

                aux_file.write("\n")
                                
                if solution.ndim == 1:
                    for k in range(grid.mx):
                        for m in range(solution.maux):
                            aux_file.write("%16.8e" % aux[k,m])
                        aux_file.write('\n')
                elif solution.ndim == 2:
                    for j in range(grid.my):
                        for k in range(grid.mx):
                            for m in range(solution.maux):
                                aux_file.write("%16.8e" % aux[k,j,m])
                            aux_file.write('\n')    
                        aux_file.write('\n')
                elif solution.ndim == 3:
                    for l in range(grid.mz):
                        for j in range(grid.my):
                            for k in range(grid.mx):
                                for m in range(solution.maux):
                                    aux_file.write("%16.8e" % aux[k,j,l,m])
                                aux_file.write('\n')
                            aux_file.write('\n')    
                        aux_file.write('\n')
    
        q_file.close()
        if solution.maux > 0 and write_aux:
            aux_file.close()
    except IOError, (errno, strerror):
        logger.error("I/O error(%s): %s" % (errno, strerror))
        raise 
    except:
        logger.error("Unexpected error:", sys.exc_info()[0])
        raise

def read_ascii(solution,frame,path='./',file_prefix='fort',read_aux=True,
                options={}):
    """read_ascii - Read in a set of files (e.g fort.qxxxx, fort.txxxx) files
    
    Calls read_ascii_t and read_ascii_q and places them into the solution 
    passed in from the arguments.
    
    Arguments:
        solution = Solution object to read the data into
        frame = Frame number to be read in
        path = Path to the current directory of the file
        file_prefix = Prefix of the files to be read in (e.g. fort.q**** where
            file_prefix='fort')
        read_aux = Boolean controlling whether or not an auxillary file will 
            try to be read in.
        options = Dictionary of options particular to this format which
            includes the following key/value pairs: 
                EMPTY
    
    """
    
    # Option parsing
    option_defaults = {}
    
    for (k,v) in option_defaults.iteritems():
        if options.has_key(k):
            exec("%s = options['%s']" % (k,k))
        else:
            exec('%s = v' % k)

    # Read in files
    read_ascii_t(solution,frame,path,file_prefix)
    read_ascii_q(solution,frame,path,file_prefix)
    if solution.grids[0].maux > 0 and read_aux:
        read_ascii_aux(solution,frame,path,file_prefix)
        
def read_ascii_t(solution,frame,path,file_prefix):
    """
    
    Reads in clawpack file fort.t000N where N = self.frameno
    
    """
    if frame < 0:
        # Don't construct file names with negative frameno values.
        raise IOError("Frame " + str(frame) + " does not exist ***")

    
    # Construct file path
    fname = os.path.join(path,)
    fname = os.path.join(path, '%s.t' % file_prefix) + str(frame).zfill(4)
    
    # Read in values from fort.t file:
    try:
        f = open(fname,'r')
        
        t = read_data_line(f)
        meqn = read_data_line(f,type='int')
        ngrids = read_data_line(f,type='int')
        ndim = read_data_line(f,type='int')
        maux = read_data_line(f,type='int')
        
        f.close()
        
        # Construct the correct number of grids with the given information
        for ng in xrange(ngrids):
            solution.create_grid(ndim=ndim)
            solution.grids[ng].t = t
            solution.grids[ng].meqn = meqn
            solution.grids[ng].maux = maux
        
    except(IOError):
        raise
    except:
        logger.error("File " + fname + " should contain t, meqn, ngrids, ndim, maux")
        raise

def read_ascii_q(solution,frame,path,file_prefix):
    """
    
    Reads in clawpack file file_prefix.q000N   where N = self.frameno
    (currently only works for ascii output files)
    
    """
    
    if frame < 0:
        # Don't construct file names with negative frameno values.
        logger.error("Frame " + str(frame) + " does not exist ***")
        raise
    
    if os.path.isfile('mapc2p.py'):
        import mapc2p
        localmapc2p = mapc2p.mapc2p
    else:
        localmapc2p = None
    
    # Open fort.q file
    fname = os.path.join(path, '%s.q' % file_prefix) + str(frame).zfill(4)
    try:
        f = open(fname,'r')
    except:
        raise IOError, "File " + fname + " does not exist ***"
    
    # Loop through every grid setting the approrpriate information
    # for ng in range(len(solution.grids)):
    for grid in solution.grids:
        if localmapc2p is not None:
            grid.mapc2p = localmapc2p
        
        grid.gridno = read_data_line(f,type='int')
        grid.level = read_data_line(f,type='int')
        
        grid.mx = read_data_line(f,type='int')
        if grid.ndim > 1:
            grid.my = read_data_line(f,type='int')
        if grid.ndim > 2:
            grid.mz = read_data_line(f,type='int')
        
        grid.xlower = read_data_line(f)
        if grid.ndim > 1:
            grid.ylower = read_data_line(f)
        if grid.ndim > 2:
            grid.zlower = read_data_line(f)
        
        grid.dx = read_data_line(f)
        if grid.ndim > 1:
            grid.dy = read_data_line(f)
        if grid.ndim > 2:
            grid.dz = read_data_line(f)
        
        blank = f.readline()
        
        # Read the data for this grid and create the grid
        grid.construct()
        
        if grid.ndim == 1:
            grid.q = empty((grid.mx,grid.meqn),'d')
            for i in range(0,grid.mx):
                l = []
                while len(l)<grid.meqn:
                    line = f.readline()
                    l = l + line.split()
                for m in range(0,grid.meqn):
                    grid.q[i,m] = float(l[m])
        
        elif grid.ndim == 2:
            grid.q = empty((grid.mx,grid.my,grid.meqn),'d')
            for j in range(grid.my):
                for i in range(grid.mx):
                    l = []
                    while len(l)<grid.meqn:
                        line = f.readline()
                        l = l + line.split()
                    for m in range(0,grid.meqn):
                        grid.q[i,j,m] = float(l[m])
                blank = f.readline()
        else:
            logger.critical("ERROR:  3D not implemented!")
            grid.q = emtpy((grid.mx,grid.my,grid.mz,grud.meqn),'d')

def read_ascii_aux(solution,frame,path,file_prefix):
    """Read in ascii auxillary file if present
    
    This function attempts to open up a file called file_prefix.aux**** or, if
    that fails, file_prefix.aux and instantiates the appropriate array in the
    solution's grid objects.
    
    Arguments:
        solution = Solution that is being read in
        frame = Current frame number
        path = Path to all the current solutions
        file_prefix = Prefix for the file name, e.g. fort
        
    Returns True if an auxillary file was successfully read, False otherwise
    """
    
    # Check for aux file
    fname1 = os.path.join(path,'%s.a' % file_prefix)+str(frame).zfill(4)
    fname2 = os.path.join(path,'%s.a' % file_prefix)
    if os.path.exists(fname1):
        fname = fname1
    elif os.path.exists(fname2):
        fname = fname2
    else:
        logger.info("Unable to open auxillary file %s or %s" % (fname1,fname2))
        return
    
    # Found a valid path, try to open and read it
    try:
        f = open(fname,'r')
    except:
        raise IOError("Could not open %s" % fname)
        
    # Read in aux file
    for n in xrange(len(solution.grids)):
        # Fetch correct grid
        gridno = read_data_line(f,type='int')
        grid = solution.grids[gridno-1]
        
        # These should match this grid already, raise exception otherwise
        if not (grid.level == read_data_line(f,type='int')):
            raise IOError("Grid level in aux file header did not match grid no %s." % grid.gridno)
        if not (grid.mx == read_data_line(f,type='int')):
            raise IOError("Grid mx in aux file header did not match grid no %s." % grid.gridno)
        if grid.ndim > 1:
            if not (grid.my == read_data_line(f,type='int')):
                raise IOError("Grid my in aux file header did not match grid no %s." % grid.gridno)
            if grid.ndim > 2:
                if not (grid.mz == read_data_line(f,type='int')):
                    raise IOError("Grid mz in aux file header did not match grid no %s." % grid.gridno)
        if not (grid.xlower == read_data_line(f,type='float')):
            raise IOError("Grid xlower in aux file header did not match grid no %s." % grid.gridno)
        if grid.ndim > 1:
            if not (grid.ylower == read_data_line(f,type='float')):
                raise IOError("Grid ylower in aux file header did not match grid no %s." % grid.gridno)
            if grid.ndim > 2:
                if not (grid.zlower == read_data_line(f,type='float')):
                    raise IOError("Grid zlower in aux file header did not match grid no %s." % grid.gridno)
        if not (grid.dx == read_data_line(f,type='float')):
            raise IOError("Grid dx in aux file header did not match grid no %s." % grid.gridno)
        if grid.ndim > 1:
            if not (grid.dy == read_data_line(f,type='float')):
                raise IOError("Grid dy in aux file header did not match grid no %s." % grid.gridno)
            if grid.ndim > 2:
                if not (grid.dz == read_data_line(f,type='float')):
                    raise IOError("Grid dz in aux file header did not match grid no %s." % grid.gridno)

        f.readline()
        
        if grid.ndim == 1:
            grid.aux = empty((grid.mx,grid.maux),'d')
            for i in range(0,grid.mx):
                l = []
                while len(l)<grid.maux:
                    line = f.readline()
                    l = l + line.split()
                for m in range(0,grid.maux):
                    grid.aux[i,m] = float(l[m])
        elif grid.ndim == 2:
            grid.aux = empty((grid.mx,grid.my,grid.maux),'d')
            for j in range(grid.my):
                for i in range(grid.mx):
                    l = []
                    while len(l)<grid.maux:
                        line = f.readline()
                        l = l + line.split()
                    for m in range(0,grid.maux):
                        grid.aux[i,j,m] = float(l[m])
                blank = f.readline()
        else:
            logger.critical("ERROR:  3D not implemented!")
            grid.aux = emtpy((grid.mx,grid.my,grid.mz,grud.maux),'d')
                    
                    
def write_fort(solution, frame, path, file_prefix, write_aux=False):
    """Legacy write_fort routine place holder
    
    This routine has been deprecated, please use write_ascii routine instead
    """
    deprecation_warning = 'DEPRECATION WARNING:  write_fort has been deprecated, please use write_ascii instead!'
    logger.warning(deprecation_warning)
    raise IOError(deprecation_warning)

def read_fort(solution,path,frame):
    """Legacy write_fort routine place holder
    
    This routine has been deprecated, please use write_ascii routine instead
    """
    deprecation_warning = 'DEPRECATION WARNING:  read_fort has been deprecated, please use read_ascii instead!'
    logger.warning(deprecation_warning)
    raise IOError(deprecation_warning)