#!/usr/bin/env python
# encoding: utf-8
#  ======================================================================
#  Package:     pyclaw.io
#  File:        hdf5.py
#  Created:     Feb 7, 2008
#  Author:      Kyle Mandli
#  ======================================================================
"""
hdf5.py

Created by Kyle Mandli on 2009-02-13.
Copyright (c) 2009 University of Washington. All rights reserved.

This module reads and writes hdf5 files via either of the following modules:
    h5py - http://code.google.com/p/h5py/
    PyTables - http://www.pytables.org/moin

It will first try h5py and then PyTables and use the correct calls
according to whichever is present on the system.  We recommend that you use
h5py as it is a minimal wrapper to the HDF5 library and will create 

To install either, you must also install the hdf5 library from the website:
    http://www.hdfgroup.org/HDF5/release/obtain5.html
    
"""

import os,sys
import logging

logger = logging.getLogger('io')

from numpy.core.multiarray import empty
from pyclaw.grid import ClawGrid
from pyclaw.util import read_data_line

# Import appropriate hdf5 package
use_h5py = False
use_PyTables = False
try:
    import h5py
    use_h5py = True
except(ImportError):
    pass
if use_h5py:
    try:
        import tables
        use_PyTables = True
    except(ImportError):
        error_msg = ("Could not import h5py or PyTables, please install " +
            "either h5py or PyTables.  See the doc_string for more " +
            "information.")
        raise Exception(error_msg)

if not use_h5py and not use_PyTables:
    logging.critical("Could not import h5py or PyTables!")

def write_hdf5(solution,frame,path,file_prefix='claw',write_aux=False,
                options={}):
    """Write out a ClawSolution to a HDF5 file.
    
    Arguments:
      solution = Pyclaw object to be output
      frame = Frame number
      path = Root path
      file_prefix = Prefix for the file name, defaults to standard 'claw'
      write_aux = Boolean controlling whether the associated auxillary array 
        should be written out.
      options = Optional argument dictionary which can contain the following
        keyword/value pairs:
            compression = (None, string ["gzip" | "lzf" | "szip"] or int 0-9) 
                Enable dataset compression. DEFLATE, LZF and (where available)
                SZIP are supported. An integer is interpreted as a GZIP level 
                for backwards compatibility
            compression_opts = (None, or special value) Setting for 
                compression filter; legal values for each filter type are:
                    "gzip"	Integer 0-9
                    "lzf"	(none allowed)
                    "szip"	2-tuple ('ec'|'nn', even integer 0-32)
                See the filters module for a detailed description of each of 
                these filters.
            chunks = (None, True or shape tuple) Store the dataset in chunked 
                format. Automatically selected if any of the other keyword 
                options are given. If you don't provide a shape tuple, the 
                library will guess one for you.
            shuffle = (True/False) Enable/disable data shuffling, which can 
                improve compression performance. Automatically enabled when 
                compression is used.
            fletcher32 = (True/False) Enable Fletcher32 error detection; may 
                be used with or without compression.
    """
    
    # Option parsing
    option_defaults = {'compression':None,'compression_opts':None,
                       'chunks':None,'shuffle':False,'fletcher32':False}
    for (k,v) in option_defaults.iteritems():
        if options.has_key(k):
            exec("%s = options['%s']" % (k,k))
        else:
            exec('%s = v' % k)
    
    # File name
    filename = os.path.join(path,'%s%s.hdf' % 
                                (file_prefix,str(frame).zfill(4)))
    
    # Write out using h5py
    if use_h5py:
        f = h5py.File(filename,'w')
        
        # For each grid, write out attributes
        for grid in solution.grids:
            # Create group for this grid
            subgroup = f.create_group('grid%s' % grid.gridno)
            
            # General grid properties
            for attr in ['t','meqn','maux','gridno','level','mx',
                            'xlower','dx','my','ylower','dy','mz','zlower',
                            'dz','mbc','mcapa','mthbc_xlower','mthbc_xupper',
                            'mthbc_ylower','mthbc_yupper','mthbc_zlower',
                            'mthbc_zupper']:
                if grid.ndim < 2 and attr in ['my','ylower','dy',
                        'mthbc_ylower','mthbc_yupper']:
                    continue
                if grid.ndim < 3 and attr in ['mz','zlower','dz',
                        'mthbc_zlower','mthbc_zupper']:
                    continue
                if eval("grid.%s is not None" % attr):
                    exec("subgroup.attrs['%s'] = grid.%s" % (attr,attr))
            
            # Write out q
            subgroup.create_dataset('q',data=grid.q,
                                        compression=compression,
                                        compression_opts=compression_opts,
                                        chunks=chunks,shuffle=shuffle,
                                        fletcher32=fletcher32)
            if write_aux and grid.maux > 0:
                subgroup.create_dataset('aux',data=grid.aux,
                                        compression=compression,
                                        compression_opts=compression_opts,
                                        chunks=chunks,shuffle=shuffle,
                                        fletcher32=fletcher32)
    
        # Flush and close the file
        f.close()
        
    # Write out using PyTables
    elif use_PyTables:
        # f = tables.openFile(filename, mode = "w", title = options['title'])
        logging.critical("PyTables has not been implemented yet.")
        raise IOError("PyTables has not been implemented yet.")
    else:
        err_msg = "No hdf5 python modules available."
        logging.critical(err_msg)
        raise Exception(err_msg)


def read_hdf5(solution,frame,path='./',file_prefix='claw',read_aux=True,
                options={}):
    """Read in a HDF5 file into a ClawSolution
    
    Arguments:
        solution = Pyclaw object to be output
        frame = Frame number
        path = Root path
        file_prefix = Prefix for the file name, defaults to 'claw'
        read_aux = Boolean controlling whether the associated auxillary array 
            should be read in.
        options = Optional argument dictionary which can contain the following
            keyword/value pairs:
                None
    """
    
    # Option parsing
    option_defaults = {}
    for (k,v) in option_defaults.iteritems():
        if options.has_key(k):
            exec("%s = options['%s']" % (k,k))
        else:
            exec('%s = v' % k)
    
    # File name
    filename = os.path.join(path,'%s%s.hdf' % 
                                (file_prefix,str(frame).zfill(4)))

    if use_h5py:
        f = h5py.File(filename,'r')
        
        for subgroup in f.iterobjects():
            # Create new grid based on shape of q variable
            ndim = len(subgroup['q'].shape) - 1
            grid = ClawGrid(ndim=ndim)
            
            # Get attributes of subgroup
            for (k,v) in subgroup.attrs.iteritems():
                exec("grid.%s = %s" % (k,v))
                    
            # Construct the rest of the grid
            grid.construct()
            
            # Read in q
            if grid.ndim == 1:
                grid.q = subgroup['q'][:,:]
            elif grid.ndim == 2:
                grid.q = subgroup['q'][:,:,:]
            elif grid.ndim == 3:
                grid.q = subgroup['q'][:,:,:,:]
            
            # Read in aux if applicable
            if read_aux and grid.maux > 0:
                if grid.ndim == 1:
                    grid.aux = subgroup['aux'][:,:]
                elif grid.ndim == 2:
                    grid.aux = subgroup['aux'][:,:,:]
                elif grid.ndim == 3:
                    grid.aux = subgroup['aux'][:,:,:,:]
                
            solution.grids.append(grid)
            
        # Flush and close the file
        f.close()
            
    elif use_PyTables:
        # f = tables.openFile(filename, mode = "r", title = options['title'])
        logging.critical("PyTables has not been implemented yet.")
        raise IOError("PyTables has not been implemented yet.")
    else:
        err_msg = "No hdf5 python modules available."
        logging.critical(err_msg)
        raise Exception(err_msg)
        