#!/usr/bin/env python
# encoding: utf-8
#  ======================================================================
#  Package:     pyclaw.io
#  File:        netcdf.py
#  Created:     Feb 17, 2009
#  Author:      Kyle Mandli
#  ======================================================================
"""
netcdf.py

Created by Kyle Mandli on 2009-02-17.
Copyright (c) 2009 University of Washington. All rights reserved.

Routines for reading and writing a NetCDF output file via either
    netcdf4-python - http://code.google.com/p/netcdf4-python/
    pupynere - http://pypi.python.org/pypi/pupynere/
These interfaces are very similar so if a different module needs to be used,
it can more than likely be inserted with a minimal of effort.

This module will first try to import the netcdf4-python module which is based
on the compiled libraries and failing that will attempt to import the pure
python interface pupynere which requires no libraries.

To install the netCDF 4 library, please see:
    http://www.unidata.ucar.edu/software/netcdf/
"""

import os,sys
import logging

from numpy.core.multiarray import empty
from pyclaw.grid import ClawGrid

logger = logging.getLogger('io')

# Import appropriate netcdf package
use_netcdf4 = False
use_pupynere = False
try:
    import netCDF4
    use_netcdf4 = True
except(ImportError):
    pass
if use_pupynere:
    try:
        import pupynere
        use_pupynere = True
    except(ImportError):
        error_msg = ("Could not import netCDF4 or Pupynere, please install " +
            "one of the available modules for netcdf files.  Refer to this " +
            "modules doc_string for more information.")
        raise Exception(error_msg)

def write_netcdf(solution,frame,path,file_prefix='claw',write_aux=False,
                    options={}):
    """Write out a NetCDF data file representation of solution
    
    Arguments:
      solution = Pyclaw object to be output
      frame = Frame number
      path = Root path
      file_prefix = Prefix for the file name, defaults to standard 'claw'
      write_aux = Boolean controlling whether the associated auxillary array 
        should be written out.
      options = Optional argument dictionary which can contain the following
        keyword/value pairs:
            description = Dictionary of key/value pairs that will be attached
                to the root group as attributes, i.e. {'time':3} 
            format = Can be one of the following netCDF flavors:
                NETCDF3_CLASSIC, NETCDF3_64BIT, NETCDF4_CLASSIC, and NETCDF4
                Default is NETCDF4
            clobber = if True (Default), file will be overwritten, if False
                an exception will be raised
            zlib = if True, data assigned to the Variable instance is 
                compressed on disk. Default False.
            complevel = the level of zlib compression to use (1 is the 
                fastest, but poorest compression, 9 is the slowest but best 
                compression). Default 6. Ignored if zlib=False.
            shuffle = if True, the HDF5 shuffle filter is applied to improve 
                compression. Default True. Ignored if zlib=False.
            fletcher32 = if True (default False), the Fletcher32 checksum 
                algorithm is used for error detection.
            contiguous = if True (default False), the variable data is stored 
                contiguously on disk. Default False. Setting to True for a 
                variable with an unlimited dimension will trigger an error.
            chunksizes = Can be used to specify the HDF5 chunksizes for each 
                dimension of the variable. A detailed discussion of HDF 
                chunking and I/O performance is available here. Basically, you
                want the chunk size for each dimension to match as closely as 
                possible the size of the data block that users will read from 
                the file. chunksizes cannot be set if contiguous=True.
            least_significant_digit = If specified, variable data will be 
                truncated (quantized). In conjunction with zlib=True this 
                produces 'lossy', but significantly more efficient 
                compression. For example, if least_significant_digit=1, data 
                will be quantized using around(scale*data)/scale, where 
                scale = 2**bits, and bits is determined so that a precision 
                of 0.1 is retained (in this case bits=4). Default is None, or 
                no quantization.
            endian = Can be used to control whether the data is stored in 
                little or big endian format on disk. Possible values are 
                little, big or native (default). The library will 
                automatically handle endian conversions when the data is read,
                but if the data is always going to be read on a computer with 
                the opposite format as the one used to create the file, there 
                may be some performance advantage to be gained by setting the 
                endian-ness.
            fill_value = If specified, the default netCDF _FillValue (the 
                value that the variable gets filled with before any data is 
                written to it) is replaced with this value. If fill_value is 
                set to False, then the variable is not pre-filled.

        * The zlib, complevel, shuffle, fletcher32, contiguous, chunksizes and
        endian keywords are silently ignored for netCDF 3 files that do not 
        use HDF5.
        
    """
    
    # Option parsing
    option_defaults = {'format':'NETCDF4','zlib':False,'complevel':6,
                       'shuffle':True,'fletcher32':False,'contiguous':False,
                       'chunksizes':None,'endian':'native',
                       'least_significant_digit':None,'fill_value':None,
                       'clobber':True,'description':{}}
    for (k,v) in option_defaults.iteritems():
        if options.has_key(k):
            exec("%s = options['%s']" % (k,k))
        else:
            exec('%s = v' % k)
            
    # Filename
    filename = os.path.join(path,"%s%s.nc" % (file_prefix,str(frame).zfill(4)))
        
    if use_netcdf4:
        # Open new file
        f = netCDF4.Dataset(filename,'w',clobber=clobber,format=format)
        
        # Loop through description dictionary and add the attributes to the
        # root group
        for (k,v) in description.iteritems():
            exec('f.%s = %s' % (k,v))
        
        # For each grid, write out attributes
        for grid in solution.grids:
            # Create group for this grid
            subgroup = f.createGroup('grid%s' % grid.gridno)
        
            # General grid properties
            for attr in ['t','meqn','maux','gridno','level','mx',
                            'xlower','dx','my','ylower','dy','mz','zlower',
                            'dz','mbc','mcapa','mthbc_xlower','mthbc_xupper',
                            'mthbc_ylower','mthbc_yupper','mthbc_zlower',
                            'mthbc_zupper']:
                if grid.ndim < 2 and attr in ['my','ylower','dy',
                        'mthbc_ylower','mthbc_yupper']:
                    continue
                if grid.ndim < 3 and attr in ['mz','zlower','dz',
                        'mthbc_zlower','mthbc_zupper']:
                    continue
                if eval("grid.%s is not None" % attr):
                    exec("subgroup.%s = grid.%s" % (attr,attr))
            
            # Create dimensions for q (and aux)
            subgroup.createDimension('mx',grid.mx)
            if grid.ndim > 1:
                subgroup.createDimension('my',grid.my)
            if grid.ndim > 2:
                subgroup.createDimension('mz',grid.mz)
            subgroup.createDimension('meqn',grid.meqn)
            
            # Write q array
            if grid.ndim == 1:
                q = subgroup.createVariable('q','f8',('mx','meqn'),zlib,
                                            complevel,shuffle,fletcher32,
                                            contiguous,chunksizes,endian,
                                            least_significant_digit,fill_value)
                q[:,:] = grid.q
            elif grid.ndim == 2:
                q = subgroup.createVariable('q','f8',('mx','my','meqn'),zlib,
                                            complevel,shuffle,fletcher32,
                                            contiguous,chunksizes,endian,
                                            least_significant_digit,fill_value)
                q[:,:,:] = grid.q
            elif grid.ndim == 3:
                q = subgroup.createVariable('q','f8',('mx','my','mz','meqn'),
                                            zlib,complevel,shuffle,fletcher32,
                                            contiguous,chunksizes,endian,
                                            least_significant_digit,fill_value)
                q[:,:,:] = grid.q
            
            # Write out aux
            if grid.maux > 0 and write_aux:
                subgroup.createDimension('maux',grid.maux)
                if grid.ndim == 1:
                    aux = subgroup.createVariable('aux','f8',('mx','maux'),
                                            zlib,complevel,shuffle,fletcher32,
                                            contiguous,chunksizes,endian,
                                            least_significant_digit,fill_value)
                    aux[:,:] = grid.aux
                elif grid.ndim == 2:
                    aux = subgroup.createVariable('aux','f8',('mx','my','maux'),
                                            zlib,complevel,shuffle,fletcher32,
                                            contiguous,chunksizes,endian,
                                            least_significant_digit,fill_value)
                    aux[:,:,:] = grid.aux
                elif grid.ndim == 3:
                    aux = subgroup.createVariable('aux','f8',('mx','my','mz','maux'),
                                            zlib,complevel,shuffle,fletcher32,
                                            contiguous,chunksizes,endian,
                                            least_significant_digit,fill_value)
                    aux[:,:,:,:] = grid.aux
            
        f.close()
    elif use_pupynere:
        logging.critical("Pupynere support has not been implemented yet.")
        raise IOError("Pupynere support has not been implemented yet.")
    else:
        err_msg = "No netcdf python modules available."
        logging.critical(err_msg)
        raise Exception(err_msg)
        
    
def read_netcdf(solution,frame,path='./',file_prefix='claw',read_aux=True,
                options={}):
    """Read in a NetCDF data files into solution
    
    Arguments:
        solution = Solution object to read the data into
        frame = Frame number to be read in
        path = Path to the current directory of the file
        file_prefix = Prefix of the files to be read in (e.g. fort.q**** where
            file_prefix='fort')
        read_aux = Boolean controlling whether or not an auxillary file will 
            try to be read in.
        options = Dictionary of options particular to this format which
            includes the following key/value pairs: 
                EMPTY
    
    """
    
    # Option parsing
    option_defaults = {}
    for (k,v) in option_defaults.iteritems():
        if options.has_key(k):
            exec("%s = options['%s']" % (k,k))
        else:
            exec('%s = v' % k)
            
    # Filename
    filename = os.path.join(path,"%s%s.nc" % (file_prefix,str(frame).zfill(4)))
        
    if use_netcdf4:
        # Open file
        f = netCDF4.Dataset(filename,'r')
        
        # We only expect subgroups of grids, otherwise we need to put some
        # sort of conditional here
        for subgroup in f.groups.itervalues():
            # Create new grid, base ndim on the shape of q
            ndim = len(subgroup.variables['q'].shape) - 1
            grid = ClawGrid(ndim=ndim)
        
            for (k,v) in subgroup.__dict__.iteritems():
                exec('grid.%s = %s' % (k,v))
                
            # Construct grid
            grid.construct()
            
            # Read in q
            if grid.ndim == 1:
                grid.q = subgroup.variables['q'][:,:]
            elif grid.ndim == 2:
                grid.q = subgroup.variables['q'][:,:,:]
            elif grid.ndim == 3:
                grid.q = subgroup.variables['q'][:,:,:,:]
            
            # Read in aux if applicable
            if read_aux and grid.maux > 0:
                if grid.ndim == 1:
                    grid.aux = subgroup.variables['aux'][:,:]
                elif grid.ndim == 2:
                    grid.aux = subgroup.variables['aux'][:,:,:]
                elif grid.ndim == 3:
                    grid.aux = subgroup.variables['aux'][:,:,:,:]
        
            solution.grids.append(grid)
            
        f.close()
    elif use_pupynere:
        logging.critical("Pupynere support has not been implemented yet.")
        raise IOError("Pupynere support has not been implemented yet.")
    else:
        err_msg = "No netcdf python modules available."
        logging.critical(err_msg)
        raise Exception(err_msg)