#!/usr/bin/env python
# encoding: utf-8
#  ======================================================================
#  Package:     Pyclaw
#  File:        ClawSolution.py
#  Created:     2008-08-07
#  Author:      Kyle Mandli - University of Washington
#  ======================================================================
"""
claw_solution module

Contains the definition of the class ClawSolution
"""
import os
import copy
import logging

from numpy.core.multiarray import empty

# Import the data readers and writers
from data import ClawGridData,ClawPlotData
from error import ClawError
from grid import ClawGrid
from io import *

class ClawSolution(object):
    """
    ClawSolution - General data structure class for all grids at one output 
        time

    Members:
        General Solution Information - 
            grids = List of grids of class ClawGrid
        
        If there is only one grid belonging to this grid, the solution will
        appear to have many of the attributes assigned to its one grid.  Some
        parameters that have in the past been parameters for all grids are
        also reachable although ClawSolution does not check to see if these
        parameters are truly universal.  The parameters included in these two
        groups are:
        
        Available only when only one grid is present:
            'mx','my','mz','xlower','ylower','zlower','xupper',
                'yupper','zupper','dx','dy','dz','mthbc_xlower',
                'mthbc_ylower','mthbc_zlower','mthbc_xupper','mthbc_yupper',
                'mthbc_zupper','q','aux','xedge','yedge','zedge','xcenter',
                'ycenter','zcenter'
        Avaiable for multiple grids:
            't','meqn','ndim','mbc','maux','mcapa'
            
        e.g. 
        >> from pylab import *
        >> from pyclaw.solution import ClawSolution
        >> ClawSolution(0)
        >> [X,Y] = meshgrid(solution.xcenter,solution.ycenter)
        >> pcolor(X,Y,solution.q[:,:,0].T)
        >> title('t = %s' % solution.t)
        >> show()
            
    Member Functions:
        __init__
        create_grid
        set_all_grids
        IO:
            write = Write this object out to the specified format
            read = Read into this object data from the given path
            plot = Plot data into current figure context
    """
    #  ======================================================================
    #  Initialization routines
    #  ======================================================================
    def __init__(self,*arg,**kargs):
        """ClawSolution Initiatlization Routine
        
        The initialization of a ClawSolution can happen these ways:
            1) args is empty and an empty ClawSolution is created
            2) args contains an object of type ClawRunData or ClawGridData
                which then instantiates this object and creates an appropriate
                grid object, appending it to the list of grids for this object
            3) args contains an object of type ClawPlotData which then
                instantiates this object, this is intended mainly for plots
            4) args is a variable number of arguments that describes the 
                location of a file to be read in to initialize the object
        
        Arguments:  The following options are valid for args 
            if args == () -> Empty ClawSolution object
            if args == ClawRunData || args == ClawGridData -> ClawSolution
                with instantiated grid with the appropriate data
            if args == ClawPlotData -> empty ClawSolution object
            if args == frame, format='ascii',path='./',file_prefix='fort'
        """

        # Viewable members of this class
        self.viewable_attributes = []
        
        # These values are set to generic values
        self.grids = []
        if len(arg) == 1:
            # Specify ClawGridData to determine basic data, note that
            # ClawRunData is a subclass of ClawGridData so either will work
            if isinstance(arg[0],ClawGridData):
                self.create_grid(data=arg[0])
            # Specify ClawPlotData as data holder
            elif isinstance(arg[0],ClawPlotData):
                pass # Not sure yet what needs to be done here yet
            # Frame number to be read in
            elif isinstance(arg[0],int): 
                frame = arg[0]
                defaults = {'format':'ascii','path':'./','file_prefix':None,
                    'read_aux':True,'options':{}}
   
                for (k,v) in defaults.iteritems():    
                    if kargs.has_key(k):
                        exec("%s = kargs['%s']" % (k,k))
                    else:
                        exec('%s = v' % k)
                self.read(frame,path,format,file_prefix,read_aux,options)
            else:
                raise ClawError("Invalid argument list for ClawSolution init!")

    #  ======================================================================
    #  Access Functions
    #  ======================================================================
    def __str__(self):
        output = "Grids:\n"
        # This is information about each of the grids
        for grid in self.grids:
            output = ''.join((output,'  %s:\n' % grid.gridno))
            output = output + str(grid)
        return str(output)
        
    def create_grid(self,data=None,ndim=None):
        """Creates a grid and appends it the grid list.
        
        A grid can be instantiated one of two ways:
          1)  A ClawRunData or ClawGridData object is passed, the grid is then
                instantiated based on that data object
          2)  Only a dimension is required and an empty grid object is 
                instantiated
        """
        
        if data is not None:
            if data.is_valid():
                self.grids.append(ClawGrid(data=data))
            else:
                raise ValueError("Invalid ClawGridData object!")
        elif ndim is not None:
            self.grids.append(ClawGrid(ndim=ndim))
        else:
            err_message = 'Invalid grid data input, need a ClawData object or a grid dictionary.'
            raise ValueError(err_message)
    
    # Override the general access methods to provide convenience to parameters
    # that are usually global.  
    def __getattribute__(self,name):
        if name in ['t','meqn','ndim','mbc','maux','mcapa']:
            return eval('self.grids[0].%s' % name)
        elif name in ['mx','my','mz','xlower','ylower','zlower','xupper',
                'yupper','zupper','dx','dy','dz','mthbc_xlower',
                'mthbc_ylower','mthbc_zlower','mthbc_xupper','mthbc_yupper',
                'mthbc_zupper','q','aux','xedge','yedge','zedge','xcenter',
                'ycenter','zcenter']:
            if len(object.__getattribute__(self,'grids')) == 1:
                return eval('self.grids[0].%s' % name)
        return object.__getattribute__(self,name)
            
    def __setattr__(self,name,value):
        if name in ['t','meqn','ndim','mbc','maux','mcapa']:
            self.set_all_grids(name,value)
        elif name in ['mx','my','mz','xlower','ylower','zlower','xupper',
                'yupper','zupper','dx','dy','dz','mthbc_xlower',
                'mthbc_ylower','mthbc_zlower','mthbc_xupper','mthbc_yupper',
                'mthbc_zupper','q','aux','xedge','yedge','zedge','xcenter',
                'ycenter','zcenter']:
            if len(object.__getattribute__(self,'grids')) == 1:
                exec('self.grids[0].%s = value' % name)
        else:
            object.__setattr__(self,name,value)
            
    def set_all_grids(self,attr,value,overwrite=True):
        """Sets all member grids attribute 'attr' to value"""
        for grid in self.grids:
            exec("change = grid.%s is None or overwrite" % attr)
            if change:
                exec("grid.%s = value" % attr)

    #  =======================================================================
    #  Copy functions
    #  =======================================================================        
    def __copy__(self):
        return self.__class__(self)
    
    def __deepcopy__(self, memo={}):
        # Create basic container
        result = self.__class__()
        result.__init__()
        
        # Populate the grids
        for grid in self.grids:
            result.grids.append(copy.deepcopy(grid))
        
        return result
    
    #  ======================================================================
    #  IO Functions
    #  ======================================================================
    def write(self,frame,path='./',format='ascii',file_prefix=None,
                write_aux=False,options={}):
        """ClawSolution.write - Writes out a representation of the solution

        Writes out a suitable representation of this solution object based on
        the format requested.  The path is built from the optional path and
        file_prefix arguments.  Will raise an IOError if unsuccessful.

        Arguments:
            frame = Frame number to append to the file output
            path - Root path, will try and create the path if it does not
                already exist.
            format - a string or list of strings containing the desired output
                formats
            file_prefix - Prefix for the file name, defaults to 'fort'
            write_aux - Write the auxillary array out as well if present
            options - Dictionary of optional arguments dependent on which
                format is being used

        """
        # Determine if we need to create the path
        if not os.path.exists(os.path.expanduser(path)):
            # Attempt to constuct the path  
            os.makedirs(path)

        # Call the correct write function based on the output format
        if isinstance(format,str):
            format_list = [format]
        elif isinstance(format,list):
            format_list = format
        # Loop over list of formats requested
        for form in format_list:
            write_func = eval('write_%s' % form)
            if file_prefix is None:
                write_func(self,frame,path,write_aux=write_aux,
                            options=options)
            else:
                write_func(self,frame,path,file_prefix=file_prefix,
                                write_aux=write_aux,options=options)
            msg = "Wrote out solution in format %s for time t=%s" % (form,self.t)
            logging.getLogger('io').info(msg)
        
        
    def read(self,frame,path='./',format='ascii',file_prefix=None,
                read_aux=False,options={}):
        """ClawSolution.read() - reads in a ClawSolution object from a file
        
        Reads in and initializes this ClawSolution with the data specified.
        This function will raise an IOError if it was unsuccessful.  

        Any format must conform to the following call signiture and return
        True if the file has been successfully read into the given solution or
        False otherwise.  Options is a dictionary of parameters that each
        format can specify.  See the ascii module for an example.
        
        read_***(solution,path,frame,file_prefix,options={})
        
        *** = format
        
        Arguments:
            frame = Frame number to be read in
            path = Base path to the files to be read
            format = Format of the file, should match on of the modules inside
                of the io package
            file_prefix = Name prefix in front of all the files, defaults to 
                whatever the format default to, e.g. fort for ascii
            options = Dictionary of optional arguments dependent on the format
                being read in
        
        Returns True if read was successful, False otherwise
        """

        read_func = eval('read_%s' % format)
        if file_prefix is None:
            read_func(self,frame,path,read_aux=read_aux,options=options)
        else:
            read_func(self,frame,path,file_prefix=file_prefix,
                                    read_aux=read_aux,options=options)
        logging.getLogger('io').info("Read in solution for time t=%s" % self.t)
        
        
    def plot(self):
        """"""
        pass
